///////////////////////////////////////////////////////////////////////////////////////////////////
//
// WinAudit Header
//
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
//
// Copyright 1987-2022 PARMAVEX SERVICES
//
// Licensed under the European Union Public Licence (EUPL), Version 1.1 or -
// as soon they will be approved by the European Commission - subsequent
// versions of the EUPL (the "Licence"). You may not use this work except in
// compliance with the Licence. You may obtain a copy of the Licence at:
//
// http://ec.europa.eu/idabc/eupl
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the Licence is distributed on an "AS IS" basis,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the Licence for the specific language governing permissions and
// limitations under the Licence. This source code is free software. It
// must not be sold, leased, rented, sub-licensed or used for any form of
// monetary recompense whatsoever. This notice must not be removed or altered
// from this source distribution.
//
///////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef WINAUDIT_WINAUDIT_H_
#define WINAUDIT_WINAUDIT_H_

///////////////////////////////////////////////////////////////////////////////////////////////////
// Remarks
///////////////////////////////////////////////////////////////////////////////////////////////////

// Use cpplint filter -whitespace/line_length

///////////////////////////////////////////////////////////////////////////////////////////////////
// Compiler directives
///////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef __cplusplus

    #define __cplusplus

#endif

///////////////////////////////////////////////////////////////////////////////////////////////////
// Includes
///////////////////////////////////////////////////////////////////////////////////////////////////

// 1. Base Library
#include "PxsBase/Header Files/PxsBase.h"

// 2. C System Files
#include <WinSock2.h>   // Winsock2 before Windows
#include <Windows.h>
#include <cfgmgr32.h>
#include <LM.h>
#include <sql.h>
#include <tchar.h>
#include <strsafe.h>    // strsafe.h after tchar.h
#include <WinDNS.h>

// 3. C++ System Files

// 4. Other Libraries

// 5. This Project
#include "WinAudit/Header Files/Resources.h"

// 6. Forwards
class Application;
class AuditRecord;
class Exception;
class Font;
class NameValue;
class String;
class StringArray;
class TreeViewItem;
class WinAuditFrame;
template< class T > class TArray;

///////////////////////////////////////////////////////////////////////////////////////////////////
// Strings
///////////////////////////////////////////////////////////////////////////////////////////////////

const wchar_t PXS_COPYRIGHT_NOTICE[]       = L"Copyright (c) Parmavex "
                                             L"Services 1987 - 2022, all "
                                             L"rights reserved.";
const wchar_t PXS_PARMAVEX_ADDRESS[]       = L"Parmavex Services, 17 Perry " 
                                             L"Hill Lane, Oldbury, Birmingham, "
                                             L"England  B68 0AG.";
const wchar_t PXS_APPLICATION_NAME[]       = L"WinAudit Freeware v3.4.3";
const wchar_t PXS_WINAUDIT_INI[]           = L"WinAudit.ini";
const wchar_t PXS_WINAUDIT_COMPUTER_GUID[] = L"WinAuditGuid";
const wchar_t PXS_REG_PATH_COMPUTER_GUID[] = L"Software\\Parmavex\\WinAudit";
const wchar_t PXS_KEYWORD_MAC_ADDRESS[]    = L"macaddress";
const wchar_t PXS_WINAUDIT_GUID_TXT[]      = L"WinAuditGuid.txt";

///////////////////////////////////////////////////////////////////////////////////////////////////
// System Management BIOS
///////////////////////////////////////////////////////////////////////////////////////////////////

// Data types
const DWORD PXS_SMBIOS_DATA_TYPE_UNKNOWN    = 0;
const DWORD PXS_SMBIOS_DATA_TYPE_BOOLEAN    = 1;    // bit
const DWORD PXS_SMBIOS_DATA_TYPE_UINT8      = 3;    // byte
const DWORD PXS_SMBIOS_DATA_TYPE_UINT16     = 5;    // word
const DWORD PXS_SMBIOS_DATA_TYPE_UINT32     = 7;    // dword
const DWORD PXS_SMBIOS_DATA_TYPE_UINT64     = 9;    // qword
const DWORD PXS_SMBIOS_DATA_TYPE_STRING     = 15;

// Type Numbers - from the specification
const BYTE  PXS_SMBIOS_TYPE_0_BIOS          = 0x00;
const BYTE  PXS_SMBIOS_TYPE_1_SYSTEM        = 0x01;
const BYTE  PXS_SMBIOS_TYPE_2_BASE_BOARD    = 0x02;
const BYTE  PXS_SMBIOS_TYPE_3_CHASSIS       = 0x03;
const BYTE  PXS_SMBIOS_TYPE_4_PROCESSOR     = 0x04;
const BYTE  PXS_SMBIOS_TYPE_5_MEM_CONTROL   = 0x05;
const BYTE  PXS_SMBIOS_TYPE_6_MEMORY_MODULE = 0x06;
const BYTE  PXS_SMBIOS_TYPE_7_CPU_CACHE     = 0x07;
const BYTE  PXS_SMBIOS_TYPE_8_PORT_CONN     = 0x08;
const BYTE  PXS_SMBIOS_TYPE_9_SYSTEM_SLOT   = 0x09;
const BYTE  PXS_SMBIOS_TYPE_16_MEM_ARRAY    = 0x10;
const BYTE  PXS_SMBIOS_TYPE_17_MEMORY_DEVICE= 0x11;

///////////////////////////////////////////////////////////////////////////////////////////////////
// Vendors
///////////////////////////////////////////////////////////////////////////////////////////////////

// Known CPU manufacturer identifiers
const DWORD PXS_CPU_MANUFACTURER_UNKNOWN     = 0;
const DWORD PXS_CPU_MANUFACTURER_INTEL       = 1;
const DWORD PXS_CPU_MANUFACTURER_AMD         = 2;

// Known HDD maufacturer name
const wchar_t PXS_HDD_MANUFACTURER_FUJITSU[] = L"Fujitsu";
const wchar_t PXS_HDD_MANUFACTURER_HITACHI[] = L"Hitachi";
const wchar_t PXS_HDD_MANUFACTURER_MAXTOR[]  = L"Maxtor";
const wchar_t PXS_HDD_MANUFACTURER_SAMSUNG[] = L"Samsung";
const wchar_t PXS_HDD_MANUFACTURER_SEAGATE[] = L"Seagate";
const wchar_t PXS_HDD_MANUFACTURER_WESTERN[] = L"Western Digital";

// Known HDD maufacturer identifiers
const DWORD PXS_HDD_MANUFACTURER_FUJITSU_ID = 1;
const DWORD PXS_HDD_MANUFACTURER_HITACHI_ID = 2;
const DWORD PXS_HDD_MANUFACTURER_MAXTOR_ID  = 3;
const DWORD PXS_HDD_MANUFACTURER_SAMSUNG_ID = 4;
const DWORD PXS_HDD_MANUFACTURER_SEAGATE_ID = 5;
const DWORD PXS_HDD_MANUFACTURER_WESTERN_ID = 6;

///////////////////////////////////////////////////////////////////////////////////////////////////
// Limits
///////////////////////////////////////////////////////////////////////////////////////////////////

// APIC
const BYTE  PXS_INVALID_APIC_ID             = 0xFF;     // = -1

// Database Login Timeout
const DWORD PXS_DB_LOGIN_TIMEOUT_SECS_DEF   = 30;
const DWORD PXS_DB_LOGIN_TIMEOUT_SECS_MIN   = 5;
const DWORD PXS_DB_LOGIN_TIMEOUT_SECS_MAX   = 99;

// Database Connection Timeout
const DWORD PXS_DB_CONNECT_TIMEOUT_SECS_DEF = 30;
const DWORD PXS_DB_CONNECT_TIMEOUT_SECS_MIN = 5;
const DWORD PXS_DB_CONNECT_TIMEOUT_SECS_MAX = 99;

// Database Query Timeout
const DWORD PXS_DB_QUERY_TIMEOUT_SECS_DEF   = 60;
const DWORD PXS_DB_QUERY_TIMEOUT_SECS_MIN   = 5;
const DWORD PXS_DB_QUERY_TIMEOUT_SECS_MAX   = 999;

// Database administration dialog
const DWORD PXS_DB_MAX_ERROR_RATE_DEFAULT   = 5;
const DWORD PXS_DB_MAX_ERROR_RATE_MIN       = 0;
const DWORD PXS_DB_MAX_ERROR_RATE_MAX       = 25;
const DWORD PXS_DB_MAX_AFFECTED_ROWS_DEFAULT= 99999;
const DWORD PXS_DB_MAX_AFFECTED_ROWS_MIN    = 1;
const DWORD PXS_DB_MAX_AFFECTED_ROWS_MAX    = 99999;
const DWORD PXS_REPORT_MAX_RECORDS_MIN      = 1;
const DWORD PXS_REPORT_MAX_RECORDS_MAX      = 99999;
const DWORD PXS_REPORT_MAX_RECORDS_DEFAULT  = 100;

///////////////////////////////////////////////////////////////////////////////////////////////////
// WinAudit Tables. Values from 1-99 are used to group data for display
// purposes. Values >= 100 are actual data tables.
///////////////////////////////////////////////////////////////////////////////////////////////////

// Interval between table numbers
const DWORD PXS_CATEGORY_INTERVAL           = 100;

// Display groups
const DWORD PXS_CATEGORY_UKNOWN             = 0;
const DWORD PXS_CATEGORY_INSTALLED_SOFTWARE = 1;
const DWORD PXS_CATEGORY_SECURITY           = 2;
const DWORD PXS_CATEGORY_GROUPS_AND_USERS   = 3;
const DWORD PXS_CATEGORY_WINDOWS_NETWORK    = 4;
const DWORD PXS_CATEGORY_NETWORK_TCPIP      = 5;
const DWORD PXS_CATEGORY_SMBIOS             = 6;
const DWORD PXS_CATEGORY_ODBC_INFORMATION   = 7;

// Computer Master
const DWORD PXS_CATEGORY_COMPUTER_MASTER    = 100;
const DWORD PXS_COMP_MASTER_COMPUTER_GUID   = PXS_CATEGORY_COMPUTER_MASTER + 1;
const DWORD PXS_COMP_MASTER_DB_USER_NAME    = PXS_CATEGORY_COMPUTER_MASTER + 2;
const DWORD PXS_COMP_MASTER_DATABASE_LOCAL  = PXS_CATEGORY_COMPUTER_MASTER + 3;
const DWORD PXS_COMP_MASTER_LAST_AUDIT_ID   = PXS_CATEGORY_COMPUTER_MASTER + 4;
const DWORD PXS_COMP_MASTER_MAC_ADDRESS     = PXS_CATEGORY_COMPUTER_MASTER + 5;
const DWORD PXS_COMP_MASTER_SMBIOS_UUID     = PXS_CATEGORY_COMPUTER_MASTER + 6;
const DWORD PXS_COMP_MASTER_ASSET_TAG       = PXS_CATEGORY_COMPUTER_MASTER + 7;
const DWORD PXS_COMP_MASTER_FQDN            = PXS_CATEGORY_COMPUTER_MASTER + 8;
const DWORD PXS_COMP_MASTER_SITE_NAME       = PXS_CATEGORY_COMPUTER_MASTER + 9;
const DWORD PXS_COMP_MASTER_DOMAIN_NAME     = PXS_CATEGORY_COMPUTER_MASTER + 10;
const DWORD PXS_COMP_MASTER_COMPUTER_NAME   = PXS_CATEGORY_COMPUTER_MASTER + 11;
const DWORD PXS_COMP_MASTER_OS_PRODUCT_ID   = PXS_CATEGORY_COMPUTER_MASTER + 12;
const DWORD PXS_COMP_MASTER_OTHER_ID        = PXS_CATEGORY_COMPUTER_MASTER + 13;
const DWORD PXS_COMP_MASTER_WINAUDIT_GUID   = PXS_CATEGORY_COMPUTER_MASTER + 14;

// Audit Master
const DWORD PXS_CATEGORY_AUDIT_MASTER       = 200;
const DWORD PXS_AUDIT_MASTER_AUDIT_GUID     = PXS_CATEGORY_AUDIT_MASTER + 1;
const DWORD PXS_AUDIT_MASTER_DB_USER_NAME   = PXS_CATEGORY_AUDIT_MASTER + 2;
const DWORD PXS_AUDIT_MASTER_DATABASE_LOCAL = PXS_CATEGORY_AUDIT_MASTER + 3;
const DWORD PXS_AUDIT_MASTER_DATABASE_UTC   = PXS_CATEGORY_AUDIT_MASTER + 4;
const DWORD PXS_AUDIT_MASTER_COMPUTER_LOCAL = PXS_CATEGORY_AUDIT_MASTER + 5;
const DWORD PXS_AUDIT_MASTER_COMPUTER_UTC   = PXS_CATEGORY_AUDIT_MASTER + 6;
const DWORD PXS_AUDIT_MASTER_COMPUTER_ID    = PXS_CATEGORY_AUDIT_MASTER + 7;

// System Overview
const DWORD PXS_CATEGORY_SYSTEM_OVERVIEW    = 300;
const DWORD PXS_SYS_OVERVIEW_COMPUTER_NAME  = PXS_CATEGORY_SYSTEM_OVERVIEW + 1;
const DWORD PXS_SYS_OVERVIEW_DOMAIN_NAME    = PXS_CATEGORY_SYSTEM_OVERVIEW + 2;
const DWORD PXS_SYS_OVERVIEW_SITE_NAME      = PXS_CATEGORY_SYSTEM_OVERVIEW + 3;
const DWORD PXS_SYS_OVERVIEW_ROLES          = PXS_CATEGORY_SYSTEM_OVERVIEW + 4;
const DWORD PXS_SYS_OVERVIEW_DESCRIPTION    = PXS_CATEGORY_SYSTEM_OVERVIEW + 5;
const DWORD PXS_SYS_OVERVIEW_OPERATING_SYS  = PXS_CATEGORY_SYSTEM_OVERVIEW + 6;
const DWORD PXS_SYS_OVERVIEW_MANUFACTURER   = PXS_CATEGORY_SYSTEM_OVERVIEW + 7;
const DWORD PXS_SYS_OVERVIEW_MODEL          = PXS_CATEGORY_SYSTEM_OVERVIEW + 8;
const DWORD PXS_SYS_OVERVIEW_SERIAL_NUMBER  = PXS_CATEGORY_SYSTEM_OVERVIEW + 9;
const DWORD PXS_SYS_OVERVIEW_ASSET_TAG      = PXS_CATEGORY_SYSTEM_OVERVIEW + 10;
const DWORD PXS_SYS_OVERVIEW_NUM_OF_PROCS   = PXS_CATEGORY_SYSTEM_OVERVIEW + 11;
const DWORD PXS_SYS_OVERVIEW_PROCESSOR_DESC = PXS_CATEGORY_SYSTEM_OVERVIEW + 12;
const DWORD PXS_SYS_OVERVIEW_TOTAL_MEMORY_MB= PXS_CATEGORY_SYSTEM_OVERVIEW + 13;
const DWORD PXS_SYS_OVERVIEW_TOTAL_HDD      = PXS_CATEGORY_SYSTEM_OVERVIEW + 14;
const DWORD PXS_SYS_OVERVIEW_DISPLAY        = PXS_CATEGORY_SYSTEM_OVERVIEW + 15;
const DWORD PXS_SYS_OVERVIEW_BIOS_VERSION   = PXS_CATEGORY_SYSTEM_OVERVIEW + 16;
const DWORD PXS_SYS_OVERVIEW_USER_ACCOUNT   = PXS_CATEGORY_SYSTEM_OVERVIEW + 17;
const DWORD PXS_SYS_OVERVIEW_SYS_UPTIME_SEC = PXS_CATEGORY_SYSTEM_OVERVIEW + 18;
const DWORD PXS_SYS_OVERVIEW_LOCAL_TIME     = PXS_CATEGORY_SYSTEM_OVERVIEW + 19;

// Installed Software - Active Setup
const DWORD PXS_CATEGORY_ACTIVE_SETUP       = 400;
const DWORD PXS_ACTIVE_SETUP_NAME           = PXS_CATEGORY_ACTIVE_SETUP + 1;
const DWORD PXS_ACTIVE_SETUP_VERSION        = PXS_CATEGORY_ACTIVE_SETUP + 2;
const DWORD PXS_ACTIVE_SETUP_INSTALLED      = PXS_CATEGORY_ACTIVE_SETUP + 3;
const DWORD PXS_ACTIVE_SETUP_INSTALL_COMMAND= PXS_CATEGORY_ACTIVE_SETUP + 4;

// Installed Software - Programs
const DWORD PXS_CATEGORY_INSTALLED_PROGS    = 500;
const DWORD PXS_INSTAL_PROGS_NAME           = PXS_CATEGORY_INSTALLED_PROGS + 1;
const DWORD PXS_INSTAL_PROGS_VENDOR         = PXS_CATEGORY_INSTALLED_PROGS + 2;
const DWORD PXS_INSTAL_PROGS_VERSION        = PXS_CATEGORY_INSTALLED_PROGS + 3;
const DWORD PXS_INSTAL_PROGS_PRODUCT_LANG   = PXS_CATEGORY_INSTALLED_PROGS + 4;
const DWORD PXS_INSTAL_PROGS_INSTALL_DATE   = PXS_CATEGORY_INSTALLED_PROGS + 5;
const DWORD PXS_INSTAL_PROGS_INSTALL_LOC    = PXS_CATEGORY_INSTALLED_PROGS + 6;
const DWORD PXS_INSTAL_PROGS_INSTALL_SRC    = PXS_CATEGORY_INSTALLED_PROGS + 7;
const DWORD PXS_INSTAL_PROGS_INSTALL_STATE  = PXS_CATEGORY_INSTALLED_PROGS + 8;
const DWORD PXS_INSTAL_PROGS_ASSIGN_TYPE    = PXS_CATEGORY_INSTALLED_PROGS + 9;
const DWORD PXS_INSTAL_PROGS_PACKAGE_CODE   = PXS_CATEGORY_INSTALLED_PROGS + 10;
const DWORD PXS_INSTAL_PROGS_PACKAGE_NAME   = PXS_CATEGORY_INSTALLED_PROGS + 11;
const DWORD PXS_INSTAL_PROGS_LOCAL_PACKAGE  = PXS_CATEGORY_INSTALLED_PROGS + 12;
const DWORD PXS_INSTAL_PROGS_PRODUCT_ID     = PXS_CATEGORY_INSTALLED_PROGS + 13;
const DWORD PXS_INSTAL_PROGS_REG_COMPANY    = PXS_CATEGORY_INSTALLED_PROGS + 14;
const DWORD PXS_INSTAL_PROGS_REG_OWNER      = PXS_CATEGORY_INSTALLED_PROGS + 15;
const DWORD PXS_INSTAL_PROGS_TIMES_USED     = PXS_CATEGORY_INSTALLED_PROGS + 16;
const DWORD PXS_INSTAL_PROGS_LAST_USED      = PXS_CATEGORY_INSTALLED_PROGS + 17;
const DWORD PXS_INSTAL_PROGS_EXE_PATH       = PXS_CATEGORY_INSTALLED_PROGS + 18;
const DWORD PXS_INSTAL_PROGS_EXE_VERSION    = PXS_CATEGORY_INSTALLED_PROGS + 19;
const DWORD PXS_INSTAL_PROGS_EXE_DESCRIP    = PXS_CATEGORY_INSTALLED_PROGS + 20;
const DWORD PXS_INSTAL_PROGS_SOFTWARE_ID    = PXS_CATEGORY_INSTALLED_PROGS + 21;

// Installed Software - Updates
const DWORD PXS_CATEGORY_SOFTWARE_UPDATES   = 600;
const DWORD PXS_SOFTWARE_UPDATES_UPDATE_ID  = PXS_CATEGORY_SOFTWARE_UPDATES + 1;
const DWORD PXS_SOFTWARE_UPDATES_INSTALL_ON = PXS_CATEGORY_SOFTWARE_UPDATES + 2;
const DWORD PXS_SOFTWARE_UPDATES_DESCRIPTION= PXS_CATEGORY_SOFTWARE_UPDATES + 3;

// Operating system
const DWORD PXS_CATEGORY_OS                 = 700;
const DWORD PXS_OS_NAME                     = PXS_CATEGORY_OS + 1;
const DWORD PXS_OS_EDITION                  = PXS_CATEGORY_OS + 2;
const DWORD PXS_OS_INSTALL_DATE             = PXS_CATEGORY_OS + 3;
const DWORD PXS_OS_REGISTERED_OWNER         = PXS_CATEGORY_OS + 4;
const DWORD PXS_OS_REGISTERED_ORG           = PXS_CATEGORY_OS + 5;
const DWORD PXS_OS_PRODUCT_ID               = PXS_CATEGORY_OS + 6;
const DWORD PXS_OS_MAJOR_VERSION_NUMBER     = PXS_CATEGORY_OS + 7;
const DWORD PXS_OS_MINOR_VERSION_NUMBER     = PXS_CATEGORY_OS + 8;
const DWORD PXS_OS_BUILD_NUMBER             = PXS_CATEGORY_OS + 9;
const DWORD PXS_OS_CSD_VERSION              = PXS_CATEGORY_OS + 10;
const DWORD PXS_OS_SERVICE_PACK_MAJOR_MINOR = PXS_CATEGORY_OS + 11;
const DWORD PXS_OS_PLUS_VERSION_NUMBER      = PXS_CATEGORY_OS + 12;
const DWORD PXS_OS_DIRECTX_VERSION          = PXS_CATEGORY_OS + 13;
const DWORD PXS_OS_WINDOWS_DIRECTORY        = PXS_CATEGORY_OS + 14;
const DWORD PXS_OS_SYSTEM_DIRECTORY         = PXS_CATEGORY_OS + 15;
const DWORD PXS_OS_TEMPORARY_DIRECTORY      = PXS_CATEGORY_OS + 16;
const DWORD PXS_OS_LANGUAGE                 = PXS_CATEGORY_OS + 17;
const DWORD PXS_OS_NUMBER_OF_BITS           = PXS_CATEGORY_OS + 18;
const DWORD PXS_OS_WIN10_VERSION            = PXS_CATEGORY_OS + 19;
const DWORD PXS_OS_WIN10_OS_BUILD           = PXS_CATEGORY_OS + 20;

// Peripherals
const DWORD PXS_CATEGORY_PERIPHERALS        = 800;
const DWORD PXS_PERIPHERALS_NAME            = PXS_CATEGORY_PERIPHERALS + 1;
const DWORD PXS_PERIPHERALS_DESCRIPTION     = PXS_CATEGORY_PERIPHERALS + 2;

// Security - Internet and Security Software - Not longer used
const DWORD PXS_CATEGORY_INET_SOFTWARE      = 900;
const DWORD PXS_INTERNET_SOFTWARE_TYPE      = PXS_CATEGORY_INET_SOFTWARE + 1;
const DWORD PXS_INTERNET_SOFTWARE_NAME      = PXS_CATEGORY_INET_SOFTWARE + 2;
const DWORD PXS_INTERNET_SOFTWARE_VERSION   = PXS_CATEGORY_INET_SOFTWARE + 3;
const DWORD PXS_INTERNET_SOFTWARE_DATA_UP   = PXS_CATEGORY_INET_SOFTWARE + 4;

// Security - Open Ports
const DWORD PXS_CATEGORY_OPEN_PORTS         = 1000;
const DWORD PXS_OPEN_PORTS_PROTOCOL         = PXS_CATEGORY_OPEN_PORTS + 1;
const DWORD PXS_OPEN_PORTS_LOCAL_ADDRESS    = PXS_CATEGORY_OPEN_PORTS + 2;
const DWORD PXS_OPEN_PORTS_LOCAL_PORT       = PXS_CATEGORY_OPEN_PORTS + 3;
const DWORD PXS_OPEN_PORTS_CAPTION          = PXS_CATEGORY_OPEN_PORTS + 4;
const DWORD PXS_OPEN_PORTS_SERVICE_NAME     = PXS_CATEGORY_OPEN_PORTS + 5;
const DWORD PXS_OPEN_PORTS_REMOTE_ADDRESS   = PXS_CATEGORY_OPEN_PORTS + 6;
const DWORD PXS_OPEN_PORTS_REMOTE_PORT      = PXS_CATEGORY_OPEN_PORTS + 7;
const DWORD PXS_OPEN_PORTS_CONNECTION_STATE = PXS_CATEGORY_OPEN_PORTS + 8;
const DWORD PXS_OPEN_PORTS_PROCESS_NAME     = PXS_CATEGORY_OPEN_PORTS + 9;
const DWORD PXS_OPEN_PORTS_PROCESS_ID       = PXS_CATEGORY_OPEN_PORTS + 10;
const DWORD PXS_OPEN_PORTS_PROCESS_DESC     = PXS_CATEGORY_OPEN_PORTS + 11;
const DWORD PXS_OPEN_PORTS_PROCESS_MANUFAC  = PXS_CATEGORY_OPEN_PORTS + 12;

// Security - Security Audit Log
const DWORD PXS_CATEGORY_SECURITY_LOG       = 1100;
const DWORD PXS_SECURITY_LOG_TIME_GENERATED = PXS_CATEGORY_SECURITY_LOG + 1;
const DWORD PXS_SECURITY_LOG_FILE           = PXS_CATEGORY_SECURITY_LOG + 2;
const DWORD PXS_SECURITY_LOG_SOURCE_NAME    = PXS_CATEGORY_SECURITY_LOG + 3;
const DWORD PXS_SECURITY_LOG_DESCRIPTION    = PXS_CATEGORY_SECURITY_LOG + 4;

// Security - Security Settings
const DWORD PXS_CATEGORY_SECUR_SETTINGS     = 1200;
const DWORD PXS_SECURITY_SETTINGS_ITEM      = PXS_CATEGORY_SECUR_SETTINGS + 1;
const DWORD PXS_SECURITY_SETTINGS_NAME      = PXS_CATEGORY_SECUR_SETTINGS + 2;
const DWORD PXS_SECURITY_SETTINGS_SETTING   = PXS_CATEGORY_SECUR_SETTINGS + 3;

// System Restore
const DWORD PXS_CATEGORY_SYSRESTORE         = 1300;
const DWORD PXS_SYSRESTORE_SEQUENCE         = PXS_CATEGORY_SYSRESTORE + 1;
const DWORD PXS_SYSRESTORE_CREATE_TIME      = PXS_CATEGORY_SYSRESTORE + 2;
const DWORD PXS_SYSRESTORE_DESCRIPTION      = PXS_CATEGORY_SYSRESTORE + 3;

// Security - User Privileges
const DWORD PXS_CATEGORY_USERPRIVS          = 1400;
const DWORD PXS_USERPRIVS_PRIVILEGE_NAME    = PXS_CATEGORY_USERPRIVS + 1;

// Security - Windows Firewall
const DWORD PXS_CATEGORY_WINDOWS_FIREWALL   = 1500;
const DWORD PXS_WINDOWS_FIREWALL_NAME       = PXS_CATEGORY_WINDOWS_FIREWALL + 1;
const DWORD PXS_WINDOWS_FIREWALL_SETTING    = PXS_CATEGORY_WINDOWS_FIREWALL + 2;

// Groups
const DWORD PXS_CATEGORY_GROUPS             = 1600;
const DWORD PXS_GROUPS_TYPE                 = PXS_CATEGORY_GROUPS + 1;
const DWORD PXS_GROUPS_NAME                 = PXS_CATEGORY_GROUPS + 2;
const DWORD PXS_GROUPS_COMMENT              = PXS_CATEGORY_GROUPS + 3;

// Group Members
const DWORD PXS_CATEGORY_GROUPMEMBERS       = 1700;
const DWORD PXS_GROUPMEMBERS_GROUP_NAME     = PXS_CATEGORY_GROUPMEMBERS + 1;
const DWORD PXS_GROUPMEMBERS_MEMBER_NAME    = PXS_CATEGORY_GROUPMEMBERS + 2;

// Group Policy Rights
const DWORD PXS_CATEGORY_GROUPPOLICY        = 1800;
const DWORD PXS_GROUPPOLICY_GROUP_NAME      = PXS_CATEGORY_GROUPPOLICY + 1;
const DWORD PXS_GROUPPOLICY_PRIVILEGE_NAME  = PXS_CATEGORY_GROUPPOLICY + 2;

// Users
const DWORD PXS_CATEGORY_USERS              = 1900;
const DWORD PXS_USERS_USERNAME              = PXS_CATEGORY_USERS + 1;
const DWORD PXS_USERS_FULL_NAME             = PXS_CATEGORY_USERS + 2;
const DWORD PXS_USERS_DESCRIPTION           = PXS_CATEGORY_USERS + 3;
const DWORD PXS_USERS_ACCOUNT_STATUS        = PXS_CATEGORY_USERS + 4;
const DWORD PXS_USERS_LOCAL_GROUPS          = PXS_CATEGORY_USERS + 5;
const DWORD PXS_USERS_GLOBAL_GROUPS         = PXS_CATEGORY_USERS + 6;
const DWORD PXS_USERS_LAST_LOGON            = PXS_CATEGORY_USERS + 7;
const DWORD PXS_USERS_LAST_LOGOFF           = PXS_CATEGORY_USERS + 8;
const DWORD PXS_USERS_NUMBER_OF_LOGONS      = PXS_CATEGORY_USERS + 9;
const DWORD PXS_USERS_BAD_PASSWORD_COUNT    = PXS_CATEGORY_USERS + 10;
const DWORD PXS_USERS_PASSWORD_AGE_DAY      = PXS_CATEGORY_USERS + 11;
const DWORD PXS_USERS_PASSWORD_EXPIRED      = PXS_CATEGORY_USERS + 12;
const DWORD PXS_USERS_ACCOUNT_EXPIRES       = PXS_CATEGORY_USERS + 13;

// Scheduled Tasks
const DWORD PXS_CATEGORY_SCHED_TASKS        = 2000;
const DWORD PXS_SCHED_TASKS_TASK_NAME       = PXS_CATEGORY_SCHED_TASKS + 1;
const DWORD PXS_SCHED_TASKS_STATUS          = PXS_CATEGORY_SCHED_TASKS + 2;
const DWORD PXS_SCHED_TASKS_SCHEDULE        = PXS_CATEGORY_SCHED_TASKS + 3;
const DWORD PXS_SCHED_TASKS_NEXT_RUN_TIME   = PXS_CATEGORY_SCHED_TASKS + 4;
const DWORD PXS_SCHED_TASKS_RUN_COMMAND     = PXS_CATEGORY_SCHED_TASKS + 5;
const DWORD PXS_SCHED_TASKS_MAX_RUN_TIME    = PXS_CATEGORY_SCHED_TASKS + 6;
const DWORD PXS_SCHED_TASKS_LAST_RUN_TIME   = PXS_CATEGORY_SCHED_TASKS + 7;
const DWORD PXS_SCHED_TASKS_LAST_RESULT     = PXS_CATEGORY_SCHED_TASKS + 8;

// Uptime Statistics
const DWORD PXS_CATEGORY_UPTIME             = 2100;
const DWORD PXS_UPTIME_DATA_START           = PXS_CATEGORY_UPTIME + 1;
const DWORD PXS_UPTIME_SYSTEM_UPTIME        = PXS_CATEGORY_UPTIME + 2;
const DWORD PXS_UPTIME_SYSTEM_AVAIL_PERCENT = PXS_CATEGORY_UPTIME + 3;
const DWORD PXS_UPTIME_TOTAL_UPTIME         = PXS_CATEGORY_UPTIME + 4;
const DWORD PXS_UPTIME_TOTAL_DOWNTIME       = PXS_CATEGORY_UPTIME + 5;
const DWORD PXS_UPTIME_TIMES_BOOTED         = PXS_CATEGORY_UPTIME + 6;
const DWORD PXS_UPTIME_CLEAN_SHUTDOWNS      = PXS_CATEGORY_UPTIME + 7;
const DWORD PXS_UPTIME_UNEXPECTED_SHUTDOWNS = PXS_CATEGORY_UPTIME + 8;

// Event Log Entries
const DWORD PXS_CATEGORY_ERROR_LOG          = 2200;
const DWORD PXS_ERROR_LOG_GENERATED         = PXS_CATEGORY_ERROR_LOG + 1;
const DWORD PXS_ERROR_LOG_FILE              = PXS_CATEGORY_ERROR_LOG + 2;
const DWORD PXS_ERROR_LOG_SOURCE_NAME       = PXS_CATEGORY_ERROR_LOG + 3;
const DWORD PXS_ERROR_LOG_DESCRIPTION       = PXS_CATEGORY_ERROR_LOG + 4;

// Windows Network - Files
const DWORD PXS_CATEGORY_WIN_NET_FILES      = 2300;
const DWORD PXS_WIN_NET_FILES_PATHNAME      = PXS_CATEGORY_WIN_NET_FILES + 1;
const DWORD PXS_WIN_NET_FILES_USERNAME      = PXS_CATEGORY_WIN_NET_FILES + 2;
const DWORD PXS_WIN_NET_FILES_LOCKS         = PXS_CATEGORY_WIN_NET_FILES + 3;
const DWORD PXS_WIN_NET_FILES_PERMISSIONS   = PXS_CATEGORY_WIN_NET_FILES + 4;

// Windows Network - Sessions
const DWORD PXS_CATEGORY_WIN_NET_SESSIONS   = 2400;
const DWORD PXS_WIN_NET_SESS_COMPUTERNAME   = PXS_CATEGORY_WIN_NET_SESSIONS + 1;
const DWORD PXS_WIN_NET_SESS_USERNAME       = PXS_CATEGORY_WIN_NET_SESSIONS + 2;
const DWORD PXS_WIN_NET_SESS_CONN_TIME_MIN  = PXS_CATEGORY_WIN_NET_SESSIONS + 3;
const DWORD PXS_WIN_NET_SESS_IDLE_TIME_MIN  = PXS_CATEGORY_WIN_NET_SESSIONS + 4;

// Windows Network - Shares
const DWORD PXS_CATEGORY_WIN_NET_SHARES     = 2500;
const DWORD PXS_WIN_NET_SHARES_NETNAME      = PXS_CATEGORY_WIN_NET_SHARES + 1;
const DWORD PXS_WIN_NET_SHARES_TYPE         = PXS_CATEGORY_WIN_NET_SHARES + 2;
const DWORD PXS_WIN_NET_SHARES_CONNECTIONS  = PXS_CATEGORY_WIN_NET_SHARES + 3;
const DWORD PXS_WIN_NET_SHARES_PATH         = PXS_CATEGORY_WIN_NET_SHARES + 4;

// Network TCP/IP Adaptors
const DWORD PXS_CATEGORY_NET_ADAPTERS       = 2600;
const DWORD PXS_NET_ADAPT_ITEM_NUMBER       = PXS_CATEGORY_NET_ADAPTERS + 1;
const DWORD PXS_NET_ADAPT_ADAPTER_NAME      = PXS_CATEGORY_NET_ADAPTERS + 2;
const DWORD PXS_NET_ADAPT_DNS_HOSTNAME      = PXS_CATEGORY_NET_ADAPTERS + 3;
const DWORD PXS_NET_ADAPT_DNS_SERVERS       = PXS_CATEGORY_NET_ADAPTERS + 4;
const DWORD PXS_NET_ADAPT_IP_ADDRESS        = PXS_CATEGORY_NET_ADAPTERS + 5;
const DWORD PXS_NET_ADAPT_IP_SUBNET         = PXS_CATEGORY_NET_ADAPTERS + 6;
const DWORD PXS_NET_ADAPT_DEFAULT_GATEWAY   = PXS_CATEGORY_NET_ADAPTERS + 7;
const DWORD PXS_NET_ADAPT_DHCP_ENABLED      = PXS_CATEGORY_NET_ADAPTERS + 8;
const DWORD PXS_NET_ADAPT_DHCP_SERVER       = PXS_CATEGORY_NET_ADAPTERS + 9;
const DWORD PXS_NET_ADAPT_DHCP_IPADDRESS    = PXS_CATEGORY_NET_ADAPTERS + 10;
const DWORD PXS_NET_ADAPT_DHCP_LEASE_OBTAIN = PXS_CATEGORY_NET_ADAPTERS + 11;
const DWORD PXS_NET_ADAPT_DHCP_LEASE_EXPIRES= PXS_CATEGORY_NET_ADAPTERS + 12;
const DWORD PXS_NET_ADAPT_STATUS_CODE       = PXS_CATEGORY_NET_ADAPTERS + 13;
const DWORD PXS_NET_ADAPT_ADAPTER_STATUS    = PXS_CATEGORY_NET_ADAPTERS + 14;
const DWORD PXS_NET_ADAPT_ADAPTER_TYPE      = PXS_CATEGORY_NET_ADAPTERS + 15;
const DWORD PXS_NET_ADAPT_MAC_ADDRESS       = PXS_CATEGORY_NET_ADAPTERS + 16;
const DWORD PXS_NET_ADAPT_CONN_STATUS       = PXS_CATEGORY_NET_ADAPTERS + 17;
const DWORD PXS_NET_ADAPT_CONNECT_MBPS      = PXS_CATEGORY_NET_ADAPTERS + 18;

// Network BIOS - Not longer used
const DWORD PXS_CATEGORY_NETBIOS            = 2700;
const DWORD PXS_NETBIOS_ITEM_NUMBER         = PXS_CATEGORY_NETBIOS + 1;
const DWORD PXS_NETBIOS_SOFTWARE_RELEASE    = PXS_CATEGORY_NETBIOS + 2;
const DWORD PXS_NETBIOS_MAC_ADDRESS         = PXS_CATEGORY_NETBIOS + 3;
const DWORD PXS_NETBIOS_ADAPTER_TYPE        = PXS_CATEGORY_NETBIOS + 4;
const DWORD PXS_NETBIOS_MAX_SESSIONS        = PXS_CATEGORY_NETBIOS + 5;
const DWORD PXS_NETBIOS_SESSIONS_PENDING    = PXS_CATEGORY_NETBIOS + 6;
const DWORD PXS_NETBIOS_MAX_DATA_BYTES      = PXS_CATEGORY_NETBIOS + 7;
const DWORD PXS_NETBIOS_MAX_DATAGRAM_BYTES  = PXS_CATEGORY_NETBIOS + 8;

// Hardware Devices
const DWORD PXS_CATEGORY_HARDWARE_DEVICES  = 2800;
const DWORD PXS_HARDWARE_DEVS_DEVICE_TYPE  = PXS_CATEGORY_HARDWARE_DEVICES + 1;
const DWORD PXS_HARDWARE_DEVS_DEVICE_NAME  = PXS_CATEGORY_HARDWARE_DEVICES + 2;
const DWORD PXS_HARDWARE_DEVS_DESCRIPTION  = PXS_CATEGORY_HARDWARE_DEVICES + 3;
const DWORD PXS_HARDWARE_DEVS_MANUFACTURER = PXS_CATEGORY_HARDWARE_DEVICES + 4;
const DWORD PXS_HARDWARE_DEVS_LOCATION     = PXS_CATEGORY_HARDWARE_DEVICES + 5;
const DWORD PXS_HARDWARE_DEVS_DRIVER_PROV  = PXS_CATEGORY_HARDWARE_DEVICES + 6;
const DWORD PXS_HARDWARE_DEVS_DRIVER_VER   = PXS_CATEGORY_HARDWARE_DEVICES + 7;
const DWORD PXS_HARDWARE_DEVS_DRIVER_DATE  = PXS_CATEGORY_HARDWARE_DEVICES + 8;
const DWORD PXS_HARDWARE_DEVS_STATUS_CODE  = PXS_CATEGORY_HARDWARE_DEVICES + 9;
const DWORD PXS_HARDWARE_DEVS_STATUS_MSG   = PXS_CATEGORY_HARDWARE_DEVICES + 10;
const DWORD PXS_HARDWARE_DEVS_CLASS_GUID   = PXS_CATEGORY_HARDWARE_DEVICES + 11;
const DWORD PXS_HARDWARE_DEVS_DEVICE_ID    = PXS_CATEGORY_HARDWARE_DEVICES + 12;

// Display Capabilities - Not longer used
const DWORD PXS_CATEGORY_DISPLAY_CAPS       = 2900;
const DWORD PXS_DISPLAY_CAPS_DISPLAY_NUMBER = PXS_CATEGORY_DISPLAY_CAPS + 1;
const DWORD PXS_DISPLAY_CAPS_TYPE           = PXS_CATEGORY_DISPLAY_CAPS + 2;
const DWORD PXS_DISPLAY_CAPS_DESCRIPTION    = PXS_CATEGORY_DISPLAY_CAPS + 3;
const DWORD PXS_DISPLAY_CAPS_CAPABLE        = PXS_CATEGORY_DISPLAY_CAPS + 4;

// Installed Printers
const DWORD PXS_CATEGORY_PRINTERS           = 3000;
const DWORD PXS_PRINTERS_ITEM_NUMBER        = PXS_CATEGORY_PRINTERS + 1;
const DWORD PXS_PRINTERS_NAME               = PXS_CATEGORY_PRINTERS + 2;
const DWORD PXS_PRINTERS_SHARE_NAME         = PXS_CATEGORY_PRINTERS + 3;
const DWORD PXS_PRINTERS_PORT_NAME          = PXS_CATEGORY_PRINTERS + 4;
const DWORD PXS_PRINTERS_LOCATION           = PXS_CATEGORY_PRINTERS + 5;
const DWORD PXS_PRINTERS_AVERAGE_PPM        = PXS_CATEGORY_PRINTERS + 6;
const DWORD PXS_PRINTERS_ATTRIBUTES         = PXS_CATEGORY_PRINTERS + 7;
const DWORD PXS_PRINTERS_PRINTER_STATUS     = PXS_CATEGORY_PRINTERS + 8;
const DWORD PXS_PRINTERS_PAPER_SIZE         = PXS_CATEGORY_PRINTERS + 9;
const DWORD PXS_PRINTERS_ORIENTATION        = PXS_CATEGORY_PRINTERS + 10;
const DWORD PXS_PRINTERS_PRINT_QUALITY      = PXS_CATEGORY_PRINTERS + 11;
const DWORD PXS_PRINTERS_DRIVER_NAME        = PXS_CATEGORY_PRINTERS + 12;
const DWORD PXS_PRINTERS_DRIVER_FILE        = PXS_CATEGORY_PRINTERS + 13;
const DWORD PXS_PRINTERS_DRIVER_VERSION     = PXS_CATEGORY_PRINTERS + 14;
const DWORD PXS_PRINTERS_DRIVER_MANUFACTURER= PXS_CATEGORY_PRINTERS + 15;
const DWORD PXS_PRINTERS_DRIVER_DESCRIPTION = PXS_CATEGORY_PRINTERS + 16;
const DWORD PXS_PRINTERS_DRIVER_DATA_FILE   = PXS_CATEGORY_PRINTERS + 17;
const DWORD PXS_PRINTERS_DRIVER_CONFIG_FILE = PXS_CATEGORY_PRINTERS + 18;

// BIOS Version
const DWORD PXS_CATEGORY_BIOS_VERSION       = 3100;
const DWORD PXS_BIOS_VERSION_BIOS_VERSION   = PXS_CATEGORY_BIOS_VERSION + 1;
const DWORD PXS_BIOS_VERSION_RELEASE_DATE   = PXS_CATEGORY_BIOS_VERSION + 2;
const DWORD PXS_BIOS_VERSION_VIDEO_BIOS_VER = PXS_CATEGORY_BIOS_VERSION + 3;
const DWORD PXS_BIOS_VERSION_VIDEO_BIOS_DATE= PXS_CATEGORY_BIOS_VERSION + 4;

// Processors - Basic Information
const DWORD PXS_CATEGORY_CPU_BASIC          = 3200;
const DWORD PXS_CPU_BASIC_ITEM_NUMBER       = PXS_CATEGORY_CPU_BASIC + 1;
const DWORD PXS_CPU_BASIC_FULL_NAME         = PXS_CATEGORY_CPU_BASIC + 2;
const DWORD PXS_CPU_BASIC_SHORT_NAME        = PXS_CATEGORY_CPU_BASIC + 3;
const DWORD PXS_CPU_BASIC_SPEED_ESTIMATE_MHZ= PXS_CATEGORY_CPU_BASIC + 4;
const DWORD PXS_CPU_BASIC_SPEED_REGISTRY_MHZ= PXS_CATEGORY_CPU_BASIC + 5;
const DWORD PXS_CPU_BASIC_TYPE              = PXS_CATEGORY_CPU_BASIC + 6;
const DWORD PXS_CPU_BASIC_MANUFACTURER      = PXS_CATEGORY_CPU_BASIC + 7;
const DWORD PXS_CPU_BASIC_SERIAL_NUMBER     = PXS_CATEGORY_CPU_BASIC + 8;
const DWORD PXS_CPU_BASIC_APIC_PHYSICAL_ID  = PXS_CATEGORY_CPU_BASIC + 9;
const DWORD PXS_CPU_BASIC_FEATURES          = PXS_CATEGORY_CPU_BASIC + 10;
const DWORD PXS_CPU_BASIC_CACHE             = PXS_CATEGORY_CPU_BASIC + 11;
const DWORD PXS_CPU_BASIC_TLB               = PXS_CATEGORY_CPU_BASIC + 12;
const DWORD PXS_CPU_BASIC_NUM_LOGICALS      = PXS_CATEGORY_CPU_BASIC + 13;

// Processors - Detailed Information - Not longer used
const DWORD PXS_CATEGORY_CPU_DETAILED       = 3300;
const DWORD PXS_CPU_DETAILED_ITEM_NUMBER    = PXS_CATEGORY_CPU_DETAILED + 1;
const DWORD PXS_CPU_DETAILED_NUM_LOGICAL    = PXS_CATEGORY_CPU_DETAILED + 2;
const DWORD PXS_CPU_DETAILED_NUM_CORES      = PXS_CATEGORY_CPU_DETAILED + 3;
const DWORD PXS_CPU_DETAILED_VENDOR_STRING  = PXS_CATEGORY_CPU_DETAILED + 4;
const DWORD PXS_CPU_DETAILED_FAMILY         = PXS_CATEGORY_CPU_DETAILED + 5;
const DWORD PXS_CPU_DETAILED_MODEL          = PXS_CATEGORY_CPU_DETAILED + 6;
const DWORD PXS_CPU_DETAILED_STEPPING       = PXS_CATEGORY_CPU_DETAILED + 7;
const DWORD PXS_CPU_DETAILED_EXT_FAMILY     = PXS_CATEGORY_CPU_DETAILED + 8;
const DWORD PXS_CPU_DETAILED_EXT_MODEL      = PXS_CATEGORY_CPU_DETAILED + 9;
const DWORD PXS_CPU_DETAILED_BRAND_ID       = PXS_CATEGORY_CPU_DETAILED + 10;
const DWORD PXS_CPU_DETAILED_FPU            = PXS_CATEGORY_CPU_DETAILED + 11;
const DWORD PXS_CPU_DETAILED_VME            = PXS_CATEGORY_CPU_DETAILED + 12;
const DWORD PXS_CPU_DETAILED_DE             = PXS_CATEGORY_CPU_DETAILED + 13;
const DWORD PXS_CPU_DETAILED_PSE            = PXS_CATEGORY_CPU_DETAILED + 14;
const DWORD PXS_CPU_DETAILED_TSC            = PXS_CATEGORY_CPU_DETAILED + 15;
const DWORD PXS_CPU_DETAILED_MSR            = PXS_CATEGORY_CPU_DETAILED + 16;
const DWORD PXS_CPU_DETAILED_PAE            = PXS_CATEGORY_CPU_DETAILED + 17;
const DWORD PXS_CPU_DETAILED_MCE            = PXS_CATEGORY_CPU_DETAILED + 18;
const DWORD PXS_CPU_DETAILED_CX8            = PXS_CATEGORY_CPU_DETAILED + 19;
const DWORD PXS_CPU_DETAILED_APIC           = PXS_CATEGORY_CPU_DETAILED + 20;
const DWORD PXS_CPU_DETAILED_SEP            = PXS_CATEGORY_CPU_DETAILED + 21;
const DWORD PXS_CPU_DETAILED_MTRR           = PXS_CATEGORY_CPU_DETAILED + 22;
const DWORD PXS_CPU_DETAILED_PGE            = PXS_CATEGORY_CPU_DETAILED + 23;
const DWORD PXS_CPU_DETAILED_MCA            = PXS_CATEGORY_CPU_DETAILED + 24;
const DWORD PXS_CPU_DETAILED_CMOV           = PXS_CATEGORY_CPU_DETAILED + 25;
const DWORD PXS_CPU_DETAILED_PAT            = PXS_CATEGORY_CPU_DETAILED + 26;
const DWORD PXS_CPU_DETAILED_PSE36          = PXS_CATEGORY_CPU_DETAILED + 27;
const DWORD PXS_CPU_DETAILED_PSN            = PXS_CATEGORY_CPU_DETAILED + 28;
const DWORD PXS_CPU_DETAILED_CLFSH          = PXS_CATEGORY_CPU_DETAILED + 29;
const DWORD PXS_CPU_DETAILED_DS             = PXS_CATEGORY_CPU_DETAILED + 30;
const DWORD PXS_CPU_DETAILED_ACPI           = PXS_CATEGORY_CPU_DETAILED + 31;
const DWORD PXS_CPU_DETAILED_MMX            = PXS_CATEGORY_CPU_DETAILED + 32;
const DWORD PXS_CPU_DETAILED_FXFR           = PXS_CATEGORY_CPU_DETAILED + 33;
const DWORD PXS_CPU_DETAILED_SSE            = PXS_CATEGORY_CPU_DETAILED + 34;
const DWORD PXS_CPU_DETAILED_SSE2           = PXS_CATEGORY_CPU_DETAILED + 35;
const DWORD PXS_CPU_DETAILED_SS             = PXS_CATEGORY_CPU_DETAILED + 36;
const DWORD PXS_CPU_DETAILED_HTT            = PXS_CATEGORY_CPU_DETAILED + 37;
const DWORD PXS_CPU_DETAILED_TM             = PXS_CATEGORY_CPU_DETAILED + 38;
const DWORD PXS_CPU_DETAILED_IA64           = PXS_CATEGORY_CPU_DETAILED + 39;
const DWORD PXS_CPU_DETAILED_PBE            = PXS_CATEGORY_CPU_DETAILED + 40;
const DWORD PXS_CPU_DETAILED_MMXEXT         = PXS_CATEGORY_CPU_DETAILED + 41;
const DWORD PXS_CPU_DETAILED_3DNOW          = PXS_CATEGORY_CPU_DETAILED + 42;
const DWORD PXS_CPU_DETAILED_3DNOWEXT       = PXS_CATEGORY_CPU_DETAILED + 43;
const DWORD PXS_CPU_DETAILED_SSE3           = PXS_CATEGORY_CPU_DETAILED + 44;
const DWORD PXS_CPU_DETAILED_EST            = PXS_CATEGORY_CPU_DETAILED + 45;
const DWORD PXS_CPU_DETAILED_CX16           = PXS_CATEGORY_CPU_DETAILED + 46;

// Processors - Cache Information - Not longer used
const DWORD PXS_CATEGORY_CPU_CACHE          = 3400;
const DWORD PXS_CPU_CACHE_LEVEL             = PXS_CATEGORY_CPU_CACHE + 1;
const DWORD PXS_CPU_CACHE_CPU_NUMBER        = PXS_CATEGORY_CPU_CACHE + 2;
const DWORD PXS_CPU_CACHE_IS_UNIFIED        = PXS_CATEGORY_CPU_CACHE + 3;
const DWORD PXS_CPU_CACHE_CODE_SIZE_KB      = PXS_CATEGORY_CPU_CACHE + 4;
const DWORD PXS_CPU_CACHE_CODE_WIDTH_BYTE   = PXS_CATEGORY_CPU_CACHE + 5;
const DWORD PXS_CPU_CACHE_CODE_ASSOC        = PXS_CATEGORY_CPU_CACHE + 6;
const DWORD PXS_CPU_CACHE_DATA_SIZE_KB      = PXS_CATEGORY_CPU_CACHE + 7;
const DWORD PXS_CPU_CACHE_DATA_WIDTH_BYTE   = PXS_CATEGORY_CPU_CACHE + 8;
const DWORD PXS_CPU_CACHE_DATA_ASSOC        = PXS_CATEGORY_CPU_CACHE + 9;
const DWORD PXS_CPU_CACHE_UNIFIED_SIZE_KB   = PXS_CATEGORY_CPU_CACHE + 10;
const DWORD PXS_CPU_CACHE_UNIFIED_WIDTH_BYTE= PXS_CATEGORY_CPU_CACHE + 11;
const DWORD PXS_CPU_CACHE_UNIFIED_ASSOC     = PXS_CATEGORY_CPU_CACHE + 12;

// Processors - Translation Look-Aside Buffer Information - Not longer used
const DWORD PXS_CATEGORY_CPU_TLB            = 3500;
const DWORD PXS_CPU_TLB_CPU_NUMBER          = PXS_CATEGORY_CPU_TLB + 1;
const DWORD PXS_CPU_TLB_IS_UNIFIED          = PXS_CATEGORY_CPU_TLB + 2;
const DWORD PXS_CPU_TLB_4K_CODE_ENTRIES     = PXS_CATEGORY_CPU_TLB + 3;
const DWORD PXS_CPU_TLB_4K_CODE_ASSOC       = PXS_CATEGORY_CPU_TLB + 4;
const DWORD PXS_CPU_TLB_4M_CODE_ENTRIES     = PXS_CATEGORY_CPU_TLB + 5;
const DWORD PXS_CPU_TLB_4M_CODE_ASSOC       = PXS_CATEGORY_CPU_TLB + 6;
const DWORD PXS_CPU_TLB_4K_DATA_ENTRIES     = PXS_CATEGORY_CPU_TLB + 7;
const DWORD PXS_CPU_TLB_4K_DATA_ASSOC       = PXS_CATEGORY_CPU_TLB + 8;
const DWORD PXS_CPU_TLB_4M_DATA_ENTRIES     = PXS_CATEGORY_CPU_TLB + 9;
const DWORD PXS_CPU_TLB_4M_DATA_ASSOC       = PXS_CATEGORY_CPU_TLB + 10;
const DWORD PXS_CPU_TLB_4K_UNIFIED_ENTRIES  = PXS_CATEGORY_CPU_TLB + 11;
const DWORD PXS_CPU_TLB_4K_UNIFIED_ASSOC    = PXS_CATEGORY_CPU_TLB + 12;
const DWORD PXS_CPU_TLB_4M_UNIFIED_ENTRIES  = PXS_CATEGORY_CPU_TLB + 13;
const DWORD PXS_CPU_TLB_4M_UNIFIED_ASSOC    = PXS_CATEGORY_CPU_TLB + 14;

// Memory Information
const DWORD PXS_CATEGORY_MEMORY             = 3600;
const DWORD PXS_MEMORY_TOTAL_RAM_MB         = PXS_CATEGORY_MEMORY + 1;
const DWORD PXS_MEMORY_FREE_RAM_MB          = PXS_CATEGORY_MEMORY + 2;
const DWORD PXS_MEMORY_MAX_SWAP_MB          = PXS_CATEGORY_MEMORY + 3;
const DWORD PXS_MEMORY_FREE_SWAP_MB         = PXS_CATEGORY_MEMORY + 4;

// Physical Disks
const DWORD PXS_CATEGORY_PHYS_DISKS         = 3700;
const DWORD PXS_PHYS_DISKS_ITEM_NUMBER      = PXS_CATEGORY_PHYS_DISKS + 1;
const DWORD PXS_PHYS_DISKS_CAPACITY_MB      = PXS_CATEGORY_PHYS_DISKS + 2;
const DWORD PXS_PHYS_DISKS_TYPE             = PXS_CATEGORY_PHYS_DISKS + 3;
const DWORD PXS_PHYS_DISKS_MANUFACTURER     = PXS_CATEGORY_PHYS_DISKS + 4;
const DWORD PXS_PHYS_DISKS_MODEL            = PXS_CATEGORY_PHYS_DISKS + 5;
const DWORD PXS_PHYS_DISKS_SERIAL_NUMBER    = PXS_CATEGORY_PHYS_DISKS + 6;
const DWORD PXS_PHYS_DISKS_FIRMWARE_REV     = PXS_CATEGORY_PHYS_DISKS + 7;
const DWORD PXS_PHYS_DISKS_CONTROL_RANK     = PXS_CATEGORY_PHYS_DISKS + 8;
const DWORD PXS_PHYS_DISKS_MASTER_SLAVE     = PXS_CATEGORY_PHYS_DISKS + 9;
const DWORD PXS_PHYS_DISKS_TOTAL_CYLINDERS  = PXS_CATEGORY_PHYS_DISKS + 10;
const DWORD PXS_PHYS_DISKS_TOTAL_HEADS      = PXS_CATEGORY_PHYS_DISKS + 11;
const DWORD PXS_PHYS_DISKS_SECTORS_PER_TRK  = PXS_CATEGORY_PHYS_DISKS + 12;
const DWORD PXS_PHYS_DISKS_BUFFER_KB        = PXS_CATEGORY_PHYS_DISKS + 13;
const DWORD PXS_PHYS_DISKS_SMART_SUPPORTED  = PXS_CATEGORY_PHYS_DISKS + 14;
const DWORD PXS_PHYS_DISKS_SMART_ENABLED    = PXS_CATEGORY_PHYS_DISKS + 15;
const DWORD PXS_PHYS_DISKS_SMART_SELF_TEST  = PXS_CATEGORY_PHYS_DISKS + 16;

// Drives
const DWORD PXS_CATEGORY_DRIVES             = 3800;
const DWORD PXS_DRIVES_LETTER               = PXS_CATEGORY_DRIVES + 1;
const DWORD PXS_DRIVES_TYPE                 = PXS_CATEGORY_DRIVES + 2;
const DWORD PXS_DRIVES_USED_SPACE_PERCENT   = PXS_CATEGORY_DRIVES + 3;
const DWORD PXS_DRIVES_USED_SPACE           = PXS_CATEGORY_DRIVES + 4;
const DWORD PXS_DRIVES_FREE_SPACE           = PXS_CATEGORY_DRIVES + 5;
const DWORD PXS_DRIVES_TOTAL_SPACE          = PXS_CATEGORY_DRIVES + 6;
const DWORD PXS_DRIVES_VOLUME_NAME          = PXS_CATEGORY_DRIVES + 7;
const DWORD PXS_DRIVES_FILE_SYSTEM          = PXS_CATEGORY_DRIVES + 8;
const DWORD PXS_DRIVES_VOLUME_SERIAL_NUMBER = PXS_CATEGORY_DRIVES + 9;
const DWORD PXS_DRIVES_SECTORS_PER_CLUSTER  = PXS_CATEGORY_DRIVES + 10;
const DWORD PXS_DRIVES_BYTES_PER_SECTOR     = PXS_CATEGORY_DRIVES + 11;
const DWORD PXS_DRIVES_FREE_CLUSTERS        = PXS_CATEGORY_DRIVES + 12;
const DWORD PXS_DRIVES_TOTAL_CLUSTERS       = PXS_CATEGORY_DRIVES + 13;

// Communication ports
const DWORD PXS_CATEGORY_COMMPORTS          = 3900;
const DWORD PXS_COMMPORTS_NUMBER            = PXS_CATEGORY_COMMPORTS + 1;
const DWORD PXS_COMMPORTS_NAME              = PXS_CATEGORY_COMMPORTS + 2;
const DWORD PXS_COMMPORTS_MONITOR_NAME      = PXS_CATEGORY_COMMPORTS + 3;
const DWORD PXS_COMMPORTS_DESCRIPTION       = PXS_CATEGORY_COMMPORTS + 4;
const DWORD PXS_COMMPORTS_TYPE              = PXS_CATEGORY_COMMPORTS + 5;

// Startup Programmes
const DWORD PXS_CATEGORY_STARTUP            = 4000;
const DWORD PXS_STARTUP_PROGRAM_NAME        = PXS_CATEGORY_STARTUP + 1;
const DWORD PXS_STARTUP_SETTINGS_FOLDER     = PXS_CATEGORY_STARTUP + 2;
const DWORD PXS_STARTUP_COMMAND             = PXS_CATEGORY_STARTUP + 3;

// NT Services
const DWORD PXS_CATEGORY_NTSERVICES         = 4100;
const DWORD PXS_NTSERVICES_SERVICE_NAME     = PXS_CATEGORY_NTSERVICES + 1;
const DWORD PXS_NTSERVICES_TYPE             = PXS_CATEGORY_NTSERVICES + 2;
const DWORD PXS_NTSERVICES_STATE            = PXS_CATEGORY_NTSERVICES + 3;
const DWORD PXS_NTSERVICES_START_MODE       = PXS_CATEGORY_NTSERVICES + 4;
const DWORD PXS_NTSERVICES_PATH_NAME        = PXS_CATEGORY_NTSERVICES + 5;
const DWORD PXS_NTSERVICES_START_NAME       = PXS_CATEGORY_NTSERVICES + 6;

// Running Processes
const DWORD PXS_CATEGORY_RUNNING_PROCS      = 4200;
const DWORD PXS_RUNNING_PROCS_NAME          = PXS_CATEGORY_RUNNING_PROCS + 1;
const DWORD PXS_RUNNING_PROCS_PROCESS_ID    = PXS_CATEGORY_RUNNING_PROCS + 2;
const DWORD PXS_RUNNING_PROCS_MEM_USAGE_KB  = PXS_CATEGORY_RUNNING_PROCS + 3;
const DWORD PXS_RUNNING_PROCS_DESCRIPTION   = PXS_CATEGORY_RUNNING_PROCS + 4;

// Loaded Modules - Not used
const DWORD PXS_CATEGORY_LOADED_MODULES     = 4300;
const DWORD PXS_LOADED_MODULES_NAME         = PXS_CATEGORY_LOADED_MODULES + 1;
const DWORD PXS_LOADED_MODULES_VERSION      = PXS_CATEGORY_LOADED_MODULES + 2;
const DWORD PXS_LOADED_MODULES_MODIFIED     = PXS_CATEGORY_LOADED_MODULES + 3;
const DWORD PXS_LOADED_MODULES_MANUFACTURER = PXS_CATEGORY_LOADED_MODULES + 4;

// System Files - Not used
const DWORD PXS_CATEGORY_SYSTEM_FILES       = 4400;
const DWORD PXS_SYSTEM_FILES_NAME           = PXS_CATEGORY_SYSTEM_FILES + 1;
const DWORD PXS_SYSTEM_FILES_VERSION        = PXS_CATEGORY_SYSTEM_FILES + 2;
const DWORD PXS_SYSTEM_FILES_MODIFIED       = PXS_CATEGORY_SYSTEM_FILES + 3;
const DWORD PXS_SYSTEM_FILES_MANUFACTURER   = PXS_CATEGORY_SYSTEM_FILES + 4;

// Find Files - Not used
const DWORD PXS_CATEGORY_FIND_FILES         = 4500;
const DWORD PXS_FIND_FILES_FILENAME         = PXS_CATEGORY_FIND_FILES + 1;
const DWORD PXS_FIND_FILES_SIZE             = PXS_CATEGORY_FIND_FILES + 2;
const DWORD PXS_FIND_FILES_MOD_TIMESTAMP    = PXS_CATEGORY_FIND_FILES + 3;
const DWORD PXS_FIND_FILES_DIRECTORY        = PXS_CATEGORY_FIND_FILES + 4;

// Display EDID and Pixel Size
const DWORD PXS_CATEGORY_DISPLAY_EDID       = 4600;
const DWORD PXS_DISPLAY_EDID_ITEM_NUMBER    = PXS_CATEGORY_DISPLAY_EDID + 1;
const DWORD PXS_DISPLAY_EDID_DISPLAY_NAME   = PXS_CATEGORY_DISPLAY_EDID + 2;
const DWORD PXS_DISPLAY_EDID_MANUFACTURER   = PXS_CATEGORY_DISPLAY_EDID + 3;
const DWORD PXS_DISPLAY_EDID_MANUFAC_DATE   = PXS_CATEGORY_DISPLAY_EDID + 4;
const DWORD PXS_DISPLAY_EDID_SERIAL_NUMBER  = PXS_CATEGORY_DISPLAY_EDID + 5;
const DWORD PXS_DISPLAY_EDID_PRODUCT_ID     = PXS_CATEGORY_DISPLAY_EDID + 6;
const DWORD PXS_DISPLAY_EDID_DISPLAY_SIZE   = PXS_CATEGORY_DISPLAY_EDID + 7;
const DWORD PXS_DISPLAY_EDID_DISPLAY_TYPE   = PXS_CATEGORY_DISPLAY_EDID + 8;
const DWORD PXS_DISPLAY_EDID_FEATURES       = PXS_CATEGORY_DISPLAY_EDID + 9;
const DWORD PXS_DISPLAY_RESOLUTION          = PXS_CATEGORY_DISPLAY_EDID +10;

// Display Adapter Data
const DWORD PXS_CATEGORY_DISPLAY_ADAPTERS  = 4700;
const DWORD PXS_DISP_ADAPT_ITEM_NUMBER     = PXS_CATEGORY_DISPLAY_ADAPTERS + 1;
const DWORD PXS_DISP_ADAPT_NAME            = PXS_CATEGORY_DISPLAY_ADAPTERS + 2;
const DWORD PXS_DISP_ADAPT_ADAPTER_RAM_MB  = PXS_CATEGORY_DISPLAY_ADAPTERS + 3;
const DWORD PXS_DISP_ADAPT_COLOR_DEPTH_BIT = PXS_CATEGORY_DISPLAY_ADAPTERS + 4;
const DWORD PXS_DISP_ADAPT_LOGPIXELSY_DPI  = PXS_CATEGORY_DISPLAY_ADAPTERS + 5;
const DWORD PXS_DISP_ADAPT_CURR_REFRESH_HZ = PXS_CATEGORY_DISPLAY_ADAPTERS + 6;
const DWORD PXS_DISP_ADAPT_VIDEO_PROCESSOR = PXS_CATEGORY_DISPLAY_ADAPTERS + 7;
const DWORD PXS_DISP_ADAPT_ADAPTER_DAC_TYPE= PXS_CATEGORY_DISPLAY_ADAPTERS + 8;
const DWORD PXS_DISP_ADAPT_ADAPTER_ID      = PXS_CATEGORY_DISPLAY_ADAPTERS + 9;
const DWORD PXS_DISP_ADAPT_BIOS            = PXS_CATEGORY_DISPLAY_ADAPTERS + 10;

// Environmental Variables
const DWORD PXS_CATEGORY_ENVIRON_VARS       = 4800;
const DWORD PXS_ENVIRON_VARS_NAME           = PXS_CATEGORY_ENVIRON_VARS + 1;
const DWORD PXS_ENVIRON_VARS_VALUE          = PXS_CATEGORY_ENVIRON_VARS + 2;

// Regional Settings
const DWORD PXS_CATEGORY_REGION_SETTINGS    = 4900;
const DWORD PXS_REGION_SETTINGS_ITEM        = PXS_CATEGORY_REGION_SETTINGS + 1;
const DWORD PXS_REGION_SETTINGS_NAME        = PXS_CATEGORY_REGION_SETTINGS + 2;
const DWORD PXS_REGION_SETTINGS_SETTING     = PXS_CATEGORY_REGION_SETTINGS + 3;

// Security Permissions
const DWORD PXS_CATEGORY_PERMISSIONS        = 5000;
const DWORD PXS_PERMISSIONS_OBJECT_NAME     = PXS_CATEGORY_PERMISSIONS + 1;
const DWORD PXS_PERMISSIONS_OBJECT_TYPE     = PXS_CATEGORY_PERMISSIONS + 2;
const DWORD PXS_PERMISSIONS_TRUSTEE         = PXS_CATEGORY_PERMISSIONS + 3;
const DWORD PXS_PERMISSIONS_ACE_TYPE        = PXS_CATEGORY_PERMISSIONS + 4;
const DWORD PXS_PERMISSIONS_PERMISSIONS     = PXS_CATEGORY_PERMISSIONS + 5;
const DWORD PXS_PERMISSIONS_ACE_FLAGS       = PXS_CATEGORY_PERMISSIONS + 6;
const DWORD PXS_PERMISSIONS_ACCESS_MASK     = PXS_CATEGORY_PERMISSIONS + 7;
const DWORD PXS_PERMISSIONS_OBJECT_OWNER    = PXS_CATEGORY_PERMISSIONS + 8;

// User Logon data
const DWORD PXS_CATEGORY_USER_LOGONS        = 5100;
const DWORD PXS_USER_LOGONS_USER_NAME       = PXS_CATEGORY_USER_LOGONS + 1;
const DWORD PXS_USER_LOGONS_FIRST_LOGON_TS  = PXS_CATEGORY_USER_LOGONS + 2;
const DWORD PXS_USER_LOGONS_LAST_LOGON_TS   = PXS_CATEGORY_USER_LOGONS + 3;
const DWORD PXS_USER_LOGONS_CONSOLE_LOGONS  = PXS_CATEGORY_USER_LOGONS + 4;
const DWORD PXS_USER_LOGONS_REMOTE_LOGONS   = PXS_CATEGORY_USER_LOGONS + 5;
const DWORD PXS_USER_LOGONS_OTHER_LOGONS    = PXS_CATEGORY_USER_LOGONS + 6;

// Software Metering
// PXS_SOFT_METER_NUMBER_STARTS was formerly "Console Starts". Remote and
// other starts are no longer used
const DWORD PXS_CATEGORY_SOFTWARE_METERING = 5200;
const DWORD PXS_SOFT_METER_FILE_NAME       = PXS_CATEGORY_SOFTWARE_METERING + 1;
const DWORD PXS_SOFT_METER_FILE_PATH       = PXS_CATEGORY_SOFTWARE_METERING + 2;
const DWORD PXS_SOFT_METER_FILE_VERSION    = PXS_CATEGORY_SOFTWARE_METERING + 3;
const DWORD PXS_SOFT_METER_PUBLISHER       = PXS_CATEGORY_SOFTWARE_METERING + 4;
const DWORD PXS_SOFT_METER_FIRST_START_TS  = PXS_CATEGORY_SOFTWARE_METERING + 5;
const DWORD PXS_SOFT_METER_LAST_START_TS   = PXS_CATEGORY_SOFTWARE_METERING + 6;
const DWORD PXS_SOFT_METER_NUMBER_STARTS   = PXS_CATEGORY_SOFTWARE_METERING + 7;
const DWORD PXS_SOFT_METER_REMOTE_STARTS   = PXS_CATEGORY_SOFTWARE_METERING + 8;
const DWORD PXS_SOFT_METER_OTHER_STARTS    = PXS_CATEGORY_SOFTWARE_METERING + 9;

// ODBC Drivers
const DWORD PXS_CATEGORY_ODBC_DRIVERS       = 5300;
const DWORD PXS_ODBC_DRIVERS_NAME           = PXS_CATEGORY_ODBC_DRIVERS + 1;
const DWORD PXS_ODBC_DRIVERS_COMPANY        = PXS_CATEGORY_ODBC_DRIVERS + 2;
const DWORD PXS_ODBC_DRIVERS_FILE_NAME      = PXS_CATEGORY_ODBC_DRIVERS + 3;
const DWORD PXS_ODBC_DRIVERS_FILE_VERSION   = PXS_CATEGORY_ODBC_DRIVERS + 4;
const DWORD PXS_ODBC_DRIVERS_DRIVER_ODBC_VER= PXS_CATEGORY_ODBC_DRIVERS + 5;

// ODBC Data Sources
const DWORD PXS_CATEGORY_ODBC_DATA_SOURCES = 5400;
const DWORD PXS_ODBC_DATA_SOURCES_DSN_TYPE = PXS_CATEGORY_ODBC_DATA_SOURCES + 1;
const DWORD PXS_ODBC_DATA_SOURCES_DSN      = PXS_CATEGORY_ODBC_DATA_SOURCES + 2;
const DWORD PXS_ODBC_DATA_SOURCES_DESC     = PXS_CATEGORY_ODBC_DATA_SOURCES + 3;

// Other/Miscellaneous Data
const DWORD PXS_CATEGORY_OTHER_AUDIT_DATA   = 5500;
const DWORD PXS_OTHER_AUDIT_DATA_ITEM_CLASS = PXS_CATEGORY_OTHER_AUDIT_DATA + 1;
const DWORD PXS_OTHER_AUDIT_DATA_ITEM_NAME  = PXS_CATEGORY_OTHER_AUDIT_DATA + 2;
const DWORD PXS_OTHER_AUDIT_DATA_ITEM_VALUE = PXS_CATEGORY_OTHER_AUDIT_DATA + 3;

// Non-Windows Executables - Not Used
const DWORD PXS_CATEGORY_NON_WINDOWS_EXE    = 5600;
const DWORD PXS_NON_WINDOWS_EXE_FILENAME    = PXS_CATEGORY_NON_WINDOWS_EXE + 1;
const DWORD PXS_NON_WINDOWS_EXE_VERSION     = PXS_CATEGORY_NON_WINDOWS_EXE + 2;
const DWORD PXS_NON_WINDOWS_EXE_MODIFIED    = PXS_CATEGORY_NON_WINDOWS_EXE + 3;
const DWORD PXS_NON_WINDOWS_EXE_MANUFAC     = PXS_CATEGORY_NON_WINDOWS_EXE + 4;
const DWORD PXS_NON_WINDOWS_EXE_DIRECTORY   = PXS_CATEGORY_NON_WINDOWS_EXE + 5;

///////////////////////////////////////////////////////////////////////////////////////////////////
// System Management BIOS Tables

// SMBIOS - Information (Type 0)
const DWORD PXS_CATEGORY_SMBIOS_INFO        = 6000;
const DWORD PXS_SMBIOS_BIOS_SMBIOS_VERSION  = PXS_CATEGORY_SMBIOS_INFO + 1;
const DWORD PXS_SMBIOS_BIOS_SMBIOS_VENDOR   = PXS_CATEGORY_SMBIOS_INFO + 2;
const DWORD PXS_SMBIOS_BIOS_VERSION         = PXS_CATEGORY_SMBIOS_INFO + 3;
const DWORD PXS_SMBIOS_BIOS_START_ADDRESS   = PXS_CATEGORY_SMBIOS_INFO + 4;
const DWORD PXS_SMBIOS_BIOS_RELEASE_DATE    = PXS_CATEGORY_SMBIOS_INFO + 5;
const DWORD PXS_SMBIOS_BIOS_ROM_SIZE_KB     = PXS_CATEGORY_SMBIOS_INFO + 6;
const DWORD PXS_SMBIOS_BIOS_CHARACTERISTICS = PXS_CATEGORY_SMBIOS_INFO + 7;
const DWORD PXS_SMBIOS_BIOS_CHAR_EXT_BYTE_1 = PXS_CATEGORY_SMBIOS_INFO + 8;
const DWORD PXS_SMBIOS_BIOS_CHAR_EXT_BYTE_2 = PXS_CATEGORY_SMBIOS_INFO + 9;
const DWORD PXS_SMBIOS_BIOS_SYS_BIOS_MAJOR  = PXS_CATEGORY_SMBIOS_INFO + 10;
const DWORD PXS_SMBIOS_BIOS_SYS_BIOS_MINOR  = PXS_CATEGORY_SMBIOS_INFO + 11;
const DWORD PXS_SMBIOS_BIOS_FIRMWARE_MAJOR  = PXS_CATEGORY_SMBIOS_INFO + 12;
const DWORD PXS_SMBIOS_BIOS_FIRMWARE_MINOR  = PXS_CATEGORY_SMBIOS_INFO + 13;
const DWORD PXS_SMBIOS_BIOS_C1_SUPPORTED    = PXS_CATEGORY_SMBIOS_INFO + 14;
const DWORD PXS_SMBIOS_BIOS_C1_ISA          = PXS_CATEGORY_SMBIOS_INFO + 15;
const DWORD PXS_SMBIOS_BIOS_C1_MCA          = PXS_CATEGORY_SMBIOS_INFO + 16;
const DWORD PXS_SMBIOS_BIOS_C1_EISA         = PXS_CATEGORY_SMBIOS_INFO + 17;
const DWORD PXS_SMBIOS_BIOS_C1_PCI          = PXS_CATEGORY_SMBIOS_INFO + 18;
const DWORD PXS_SMBIOS_BIOS_C1_PCMCIA       = PXS_CATEGORY_SMBIOS_INFO + 19;
const DWORD PXS_SMBIOS_BIOS_C1_PNP          = PXS_CATEGORY_SMBIOS_INFO + 20;
const DWORD PXS_SMBIOS_BIOS_C1_APM          = PXS_CATEGORY_SMBIOS_INFO + 21;
const DWORD PXS_SMBIOS_BIOS_C1_FLASH        = PXS_CATEGORY_SMBIOS_INFO + 22;
const DWORD PXS_SMBIOS_BIOS_C1_SHADOW       = PXS_CATEGORY_SMBIOS_INFO + 23;
const DWORD PXS_SMBIOS_BIOS_C1_VL_VESA      = PXS_CATEGORY_SMBIOS_INFO + 24;
const DWORD PXS_SMBIOS_BIOS_C1_SUPPORT_ESCD = PXS_CATEGORY_SMBIOS_INFO + 25;
const DWORD PXS_SMBIOS_BIOS_C1_BOOT_CD      = PXS_CATEGORY_SMBIOS_INFO + 26;
const DWORD PXS_SMBIOS_BIOS_C1_SELECT_BOOT  = PXS_CATEGORY_SMBIOS_INFO + 27;
const DWORD PXS_SMBIOS_BIOS_C1_ROM_SOCKETED = PXS_CATEGORY_SMBIOS_INFO + 28;
const DWORD PXS_SMBIOS_BIOS_C1_BOOT_PCMCIA  = PXS_CATEGORY_SMBIOS_INFO + 29;
const DWORD PXS_SMBIOS_BIOS_C1_EDD          = PXS_CATEGORY_SMBIOS_INFO + 30;
const DWORD PXS_SMBIOS_BIOS_C1_FDD_NEC_1_2MB= PXS_CATEGORY_SMBIOS_INFO + 31;
const DWORD PXS_SMBIOS_BIOS_C1_FDD_TOS_1_2MB= PXS_CATEGORY_SMBIOS_INFO + 32;
const DWORD PXS_SMBIOS_BIOS_C1_FDD_360KB    = PXS_CATEGORY_SMBIOS_INFO + 33;
const DWORD PXS_SMBIOS_BIOS_C1_FDD_1_2MB    = PXS_CATEGORY_SMBIOS_INFO + 34;
const DWORD PXS_SMBIOS_BIOS_C1_FDD_720KB    = PXS_CATEGORY_SMBIOS_INFO + 35;
const DWORD PXS_SMBIOS_BIOS_C1_FDD_2_88MB   = PXS_CATEGORY_SMBIOS_INFO + 36;
const DWORD PXS_SMBIOS_BIOS_C1_PRINT_SCREEN = PXS_CATEGORY_SMBIOS_INFO + 37;
const DWORD PXS_SMBIOS_BIOS_C1_KEYBOARD_8042= PXS_CATEGORY_SMBIOS_INFO + 38;
const DWORD PXS_SMBIOS_BIOS_C1_SERIAL       = PXS_CATEGORY_SMBIOS_INFO + 39;
const DWORD PXS_SMBIOS_BIOS_C1_PRINTER      = PXS_CATEGORY_SMBIOS_INFO + 40;
const DWORD PXS_SMBIOS_BIOS_C1_CGA_MONO     = PXS_CATEGORY_SMBIOS_INFO + 41;
const DWORD PXS_SMBIOS_BIOS_C1_NEC_PC98     = PXS_CATEGORY_SMBIOS_INFO + 42;
const DWORD PXS_SMBIOS_BIOS_CX1_SUPPORT_ACPI= PXS_CATEGORY_SMBIOS_INFO + 43;
const DWORD PXS_SMBIOS_BIOS_CX1_LEGACY_USB  = PXS_CATEGORY_SMBIOS_INFO + 44;
const DWORD PXS_SMBIOS_BIOS_CX1_SUPPORT_AGP = PXS_CATEGORY_SMBIOS_INFO + 45;
const DWORD PXS_SMBIOS_BIOS_CX1_BOOT_I2O    = PXS_CATEGORY_SMBIOS_INFO + 46;
const DWORD PXS_SMBIOS_BIOS_CX1_BOOT_LS120  = PXS_CATEGORY_SMBIOS_INFO + 47;
const DWORD PXS_SMBIOS_BIOS_CX1_BOOT_ATAPI_Z= PXS_CATEGORY_SMBIOS_INFO + 48;
const DWORD PXS_SMBIOS_BIOS_CX1_BOOT_1394   = PXS_CATEGORY_SMBIOS_INFO + 49;
const DWORD PXS_SMBIOS_BIOS_CX1_SMART_BAT   = PXS_CATEGORY_SMBIOS_INFO + 50;

// SMBIOS - System Information (Type 1)
const DWORD PXS_CATEGORY_SMBIOS_SYSINFO     = 6100;
const DWORD PXS_SMBIOS_SYSINFO_MANUFACTURER = PXS_CATEGORY_SMBIOS_SYSINFO + 1;
const DWORD PXS_SMBIOS_SYSINFO_PRODUCT_NAME = PXS_CATEGORY_SMBIOS_SYSINFO + 2;
const DWORD PXS_SMBIOS_SYSINFO_VERSION      = PXS_CATEGORY_SMBIOS_SYSINFO + 3;
const DWORD PXS_SMBIOS_SYSINFO_SERIAL_NUMBER= PXS_CATEGORY_SMBIOS_SYSINFO + 4;
const DWORD PXS_SMBIOS_SYSINFO_UUID         = PXS_CATEGORY_SMBIOS_SYSINFO + 5;
const DWORD PXS_SMBIOS_SYSINFO_WAKEUP_TYPE  = PXS_CATEGORY_SMBIOS_SYSINFO + 6;
const DWORD PXS_SMBIOS_SYSINFO_SKU_NUMBER   = PXS_CATEGORY_SMBIOS_SYSINFO + 7;
const DWORD PXS_SMBIOS_SYSINFO_FAMILY       = PXS_CATEGORY_SMBIOS_SYSINFO + 8;

// SMBIOS - Base Board Information (Type 2)
const DWORD PXS_CATEGORY_SMBIOS_BOARD       = 6200;
const DWORD PXS_SMBIOS_BOARD_ITEM_NUMBER    = PXS_CATEGORY_SMBIOS_BOARD + 1;
const DWORD PXS_SMBIOS_BOARD_MANUFACTURER   = PXS_CATEGORY_SMBIOS_BOARD + 2;
const DWORD PXS_SMBIOS_BOARD_PRODUCT        = PXS_CATEGORY_SMBIOS_BOARD + 3;
const DWORD PXS_SMBIOS_BOARD_VERSION        = PXS_CATEGORY_SMBIOS_BOARD + 4;
const DWORD PXS_SMBIOS_BOARD_SERIAL_NUMBER  = PXS_CATEGORY_SMBIOS_BOARD + 5;
const DWORD PXS_SMBIOS_BOARD_ASSET_TAG      = PXS_CATEGORY_SMBIOS_BOARD + 6;
const DWORD PXS_SMBIOS_BOARD_FEATURES_FLAG  = PXS_CATEGORY_SMBIOS_BOARD + 7;
const DWORD PXS_SMBIOS_BOARD_BOARD_TYPE     = PXS_CATEGORY_SMBIOS_BOARD + 8;

// SMBIOS - Chassis Information (Type 3)
const DWORD PXS_CATEGORY_SMBIOS_CHASSIS     = 6300;
const DWORD PXS_SMBIOS_CHASSIS_MANUFACTURER = PXS_CATEGORY_SMBIOS_CHASSIS + 1;
const DWORD PXS_SMBIOS_CHASSIS_TYPE         = PXS_CATEGORY_SMBIOS_CHASSIS + 2;
const DWORD PXS_SMBIOS_CHASSIS_VERSION      = PXS_CATEGORY_SMBIOS_CHASSIS + 3;
const DWORD PXS_SMBIOS_CHASSIS_SERIAL_NUMBER= PXS_CATEGORY_SMBIOS_CHASSIS + 4;
const DWORD PXS_SMBIOS_CHASSIS_ASSET_TAG_NUM= PXS_CATEGORY_SMBIOS_CHASSIS + 5;
const DWORD PXS_SMBIOS_CHASSIS_BOOT_UP_STATE= PXS_CATEGORY_SMBIOS_CHASSIS + 6;
const DWORD PXS_SMBIOS_CHASSIS_POWER_SUPPLY = PXS_CATEGORY_SMBIOS_CHASSIS + 7;
const DWORD PXS_SMBIOS_CHASSIS_THERMAL_STATE= PXS_CATEGORY_SMBIOS_CHASSIS + 8;
const DWORD PXS_SMBIOS_CHASSIS_SECUR_STATUS = PXS_CATEGORY_SMBIOS_CHASSIS + 9;
const DWORD PXS_SMBIOS_CHASSIS_OEM_DEFINED  = PXS_CATEGORY_SMBIOS_CHASSIS + 10;
const DWORD PXS_SMBIOS_CHASSIS_HEIGHT_U     = PXS_CATEGORY_SMBIOS_CHASSIS + 11;
const DWORD PXS_SMBIOS_CHASSIS_POWER_CORDS  = PXS_CATEGORY_SMBIOS_CHASSIS + 12;

// SMBIOS - Processor Information (Type 4)
const DWORD PXS_CATEGORY_SMBIOS_PROC        = 6400;
const DWORD PXS_SMBIOS_PROCESSOR_ITEM_NUMBER= PXS_CATEGORY_SMBIOS_PROC + 1;
const DWORD PXS_SMBIOS_PROCESSOR_SOCKET     = PXS_CATEGORY_SMBIOS_PROC + 2;
const DWORD PXS_SMBIOS_PROCESSOR_TYPE       = PXS_CATEGORY_SMBIOS_PROC + 3;
const DWORD PXS_SMBIOS_PROCESSOR_FAMILY     = PXS_CATEGORY_SMBIOS_PROC + 4;
const DWORD PXS_SMBIOS_PROCESSOR_MANUFAC    = PXS_CATEGORY_SMBIOS_PROC + 5;
const DWORD PXS_SMBIOS_PROCESSOR_ID         = PXS_CATEGORY_SMBIOS_PROC + 6;
const DWORD PXS_SMBIOS_PROCESSOR_VERSION    = PXS_CATEGORY_SMBIOS_PROC + 7;
const DWORD PXS_SMBIOS_PROCESSOR_VOLTAGE    = PXS_CATEGORY_SMBIOS_PROC + 8;
const DWORD PXS_SMBIOS_PROCESSOR_EXT_CLK_MHZ= PXS_CATEGORY_SMBIOS_PROC + 9;
const DWORD PXS_SMBIOS_PROCESSOR_MAXIMUM_MHZ= PXS_CATEGORY_SMBIOS_PROC + 10;
const DWORD PXS_SMBIOS_PROCESSOR_CURRENT_MHZ= PXS_CATEGORY_SMBIOS_PROC + 11;
const DWORD PXS_SMBIOS_PROCESSOR_STATUS     = PXS_CATEGORY_SMBIOS_PROC + 12;
const DWORD PXS_SMBIOS_PROCESSOR_UPGRADE    = PXS_CATEGORY_SMBIOS_PROC + 13;
const DWORD PXS_SMBIOS_PROCESSOR_SERIAL_NUM = PXS_CATEGORY_SMBIOS_PROC + 14;
const DWORD PXS_SMBIOS_PROCESSOR_ASSET_TAG  = PXS_CATEGORY_SMBIOS_PROC + 15;
const DWORD PXS_SMBIOS_PROCESSOR_PART_NUM   = PXS_CATEGORY_SMBIOS_PROC + 16;
const DWORD PXS_SMBIOS_PROCESSOR_CORE_COUNT = PXS_CATEGORY_SMBIOS_PROC + 17;
const DWORD PXS_SMBIOS_PROCESSOR_CORE_ENABLE= PXS_CATEGORY_SMBIOS_PROC + 18;
const DWORD PXS_SMBIOS_PROCESSOR_THREADS    = PXS_CATEGORY_SMBIOS_PROC + 19;
const DWORD PXS_SMBIOS_PROCESSOR_CHARACTERIS= PXS_CATEGORY_SMBIOS_PROC + 20;
const DWORD PXS_SMBIOS_PROCESSOR_FAMILY_2   = PXS_CATEGORY_SMBIOS_PROC + 21;

// SMBIOS - Memory Controller Information (Type 5)
const DWORD PXS_CATEGORY_SMBIOS_MEMCTRL     = 6500;
const DWORD PXS_SMBIOS_MEMCTRL_ITEM_NUMBER  = PXS_CATEGORY_SMBIOS_MEMCTRL + 1;
const DWORD PXS_SMBIOS_MEMCTRL_ERR_DETECT   = PXS_CATEGORY_SMBIOS_MEMCTRL + 2;
const DWORD PXS_SMBIOS_MEMCTRL_ERR_CORRECT  = PXS_CATEGORY_SMBIOS_MEMCTRL + 3;
const DWORD PXS_SMBIOS_MEMCTRL_SUPPORT_INTER= PXS_CATEGORY_SMBIOS_MEMCTRL + 4;
const DWORD PXS_SMBIOS_MEMCTRL_CURRENT_INTER= PXS_CATEGORY_SMBIOS_MEMCTRL + 5;
const DWORD PXS_SMBIOS_MEMCTRL_MAX_MODULE_MB= PXS_CATEGORY_SMBIOS_MEMCTRL + 6;
const DWORD PXS_SMBIOS_MEMCTRL_SPEEDS       = PXS_CATEGORY_SMBIOS_MEMCTRL + 7;
const DWORD PXS_SMBIOS_MEMCTRL_MEM_TYPES    = PXS_CATEGORY_SMBIOS_MEMCTRL + 8;
const DWORD PXS_SMBIOS_MEMCTRL_VOLTAGE      = PXS_CATEGORY_SMBIOS_MEMCTRL + 9;
const DWORD PXS_SMBIOS_MEMCTRL_NUM_SLOTS    = PXS_CATEGORY_SMBIOS_MEMCTRL + 10;

// SMBIOS - Memory Module Information (Type 6)
const DWORD PXS_CATEGORY_SMBIOS_MEMMODULE   = 6600;
const DWORD PXS_SMBIOS_MEMMODULE_ITEM_NUMBER= PXS_CATEGORY_SMBIOS_MEMMODULE + 1;
const DWORD PXS_SMBIOS_MEMMODULE_SOCKET     = PXS_CATEGORY_SMBIOS_MEMMODULE + 2;
const DWORD PXS_SMBIOS_MEMMODULE_BANK_CONN  = PXS_CATEGORY_SMBIOS_MEMMODULE + 3;
const DWORD PXS_SMBIOS_MEMMODULE_SPEED_NS   = PXS_CATEGORY_SMBIOS_MEMMODULE + 4;
const DWORD PXS_SMBIOS_MEMMODULE_MEM_TYPE   = PXS_CATEGORY_SMBIOS_MEMMODULE + 5;
const DWORD PXS_SMBIOS_MEMMODULE_INSTAL_SIZE= PXS_CATEGORY_SMBIOS_MEMMODULE + 6;
const DWORD PXS_SMBIOS_MEMMODULE_ENABLE_SIZE= PXS_CATEGORY_SMBIOS_MEMMODULE + 7;
const DWORD PXS_SMBIOS_MEMMODULE_ERR_STATUS = PXS_CATEGORY_SMBIOS_MEMMODULE + 8;

// SMBIOS - CPU Cache Information (Type 7)
const DWORD PXS_CATEGORY_SMBIOS_CPUCACHE    = 6700;
const DWORD PXS_SMBIOS_CPUCACHE_ITEM_NUMBER = PXS_CATEGORY_SMBIOS_CPUCACHE + 1;
const DWORD PXS_SMBIOS_CPUCACHE_SOCKET      = PXS_CATEGORY_SMBIOS_CPUCACHE + 2;
const DWORD PXS_SMBIOS_CPUCACHE_CONFIG      = PXS_CATEGORY_SMBIOS_CPUCACHE + 3;
const DWORD PXS_SMBIOS_CPUCACHE_MAX_SIZE_KB = PXS_CATEGORY_SMBIOS_CPUCACHE + 4;
const DWORD PXS_SMBIOS_CPUCACHE_INST_SIZE_KB= PXS_CATEGORY_SMBIOS_CPUCACHE + 5;
const DWORD PXS_SMBIOS_CPUCACHE_SUPPORT_SRAM= PXS_CATEGORY_SMBIOS_CPUCACHE + 6;
const DWORD PXS_SMBIOS_CPUCACHE_CURRENT_SRAM= PXS_CATEGORY_SMBIOS_CPUCACHE + 7;
const DWORD PXS_SMBIOS_CPUCACHE_SPEED_NS    = PXS_CATEGORY_SMBIOS_CPUCACHE + 8;
const DWORD PXS_SMBIOS_CPUCACHE_ERR_CORRECT = PXS_CATEGORY_SMBIOS_CPUCACHE + 9;
const DWORD PXS_SMBIOS_CPUCACHE_SYSTEM_TYPE = PXS_CATEGORY_SMBIOS_CPUCACHE + 10;
const DWORD PXS_SMBIOS_CPUCACHE_ASSOCIATVITY= PXS_CATEGORY_SMBIOS_CPUCACHE + 11;

// SMBIOS - Port Connector Information (Type 8)
const DWORD PXS_CATEGORY_SMBIOS_PORTCONN    = 6800;
const DWORD PXS_SMBIOS_PORTCONN_ITEM_NUMBER = PXS_CATEGORY_SMBIOS_PORTCONN + 1;
const DWORD PXS_SMBIOS_PORTCONN_INT_REF     = PXS_CATEGORY_SMBIOS_PORTCONN + 2;
const DWORD PXS_SMBIOS_PORTCONN_INT_CONN    = PXS_CATEGORY_SMBIOS_PORTCONN + 3;
const DWORD PXS_SMBIOS_PORTCONN_EXT_REF     = PXS_CATEGORY_SMBIOS_PORTCONN + 4;
const DWORD PXS_SMBIOS_PORTCONN_EXT_CONN    = PXS_CATEGORY_SMBIOS_PORTCONN + 5;
const DWORD PXS_SMBIOS_PORTCONN_PORT_TYPE   = PXS_CATEGORY_SMBIOS_PORTCONN + 6;

// SMBIOS - System Slot Information (Type 9)
const DWORD PXS_CATEGORY_SMBIOS_SYSSLOT     = 6900;
const DWORD PXS_SMBIOS_SYSSLOT_ITEM_NUMBER  = PXS_CATEGORY_SMBIOS_SYSSLOT + 1;
const DWORD PXS_SMBIOS_SYSSLOT_DESIGNATION  = PXS_CATEGORY_SMBIOS_SYSSLOT + 2;
const DWORD PXS_SMBIOS_SYSSLOT_SLOT_TYPE    = PXS_CATEGORY_SMBIOS_SYSSLOT + 3;
const DWORD PXS_SMBIOS_SYSSLOT_BUS_WIDTH    = PXS_CATEGORY_SMBIOS_SYSSLOT + 4;
const DWORD PXS_SMBIOS_SYSSLOT_CURRENT_USAGE= PXS_CATEGORY_SMBIOS_SYSSLOT + 5;
const DWORD PXS_SMBIOS_SYSSLOT_SLOT_LENGTH  = PXS_CATEGORY_SMBIOS_SYSSLOT + 6;
const DWORD PXS_SMBIOS_SYSSLOT_SLOT_ID      = PXS_CATEGORY_SMBIOS_SYSSLOT + 7;
const DWORD PXS_SMBIOS_SYSSLOT_C_1          = PXS_CATEGORY_SMBIOS_SYSSLOT + 8;
const DWORD PXS_SMBIOS_SYSSLOT_C_2          = PXS_CATEGORY_SMBIOS_SYSSLOT + 9;
const DWORD PXS_SMBIOS_SYSSLOT_SEGMENT_NUM  = PXS_CATEGORY_SMBIOS_SYSSLOT + 10;
const DWORD PXS_SMBIOS_SYSSLOT_BUS_NUM      = PXS_CATEGORY_SMBIOS_SYSSLOT + 11;
const DWORD PXS_SMBIOS_SYSSLOT_DEVICE_NUM   = PXS_CATEGORY_SMBIOS_SYSSLOT + 12;

// SMBIOS - Physical Memory Array Information (Type 16)
const DWORD PXS_CATEGORY_SMBIOS_MEMARRAY    = 7000;
const DWORD PXS_SMBIOS_MEMARRAY_ITEM_NUMBER = PXS_CATEGORY_SMBIOS_MEMARRAY + 1;
const DWORD PXS_SMBIOS_MEMARRAY_LOCATION    = PXS_CATEGORY_SMBIOS_MEMARRAY + 2;
const DWORD PXS_SMBIOS_MEMARRAY_USE         = PXS_CATEGORY_SMBIOS_MEMARRAY + 3;
const DWORD PXS_SMBIOS_MEMARRAY_ERR_CORRECT = PXS_CATEGORY_SMBIOS_MEMARRAY + 4;
const DWORD PXS_SMBIOS_MEMARRAY_MAX_CAP_KB  = PXS_CATEGORY_SMBIOS_MEMARRAY + 5;
const DWORD PXS_SMBIOS_MEMARRAY_ERROR_HANDLE= PXS_CATEGORY_SMBIOS_MEMARRAY + 6;
const DWORD PXS_SMBIOS_MEMARRAY_NUM_DEVICES = PXS_CATEGORY_SMBIOS_MEMARRAY + 7;
const DWORD PXS_SMBIOS_MEMARRAY_EX_MAX_CAP  = PXS_CATEGORY_SMBIOS_MEMARRAY + 8;

// SMBIOS - Memory Device Information (Type 17)
const DWORD PXS_CATEGORY_SMBIOS_MEMDEV      = 7100;
const DWORD PXS_SMBIOS_MEM_DEV_ITEM_NUMBER  = PXS_CATEGORY_SMBIOS_MEMDEV + 1;
const DWORD PXS_SMBIOS_MEM_DEV_ERROR_HANDLE = PXS_CATEGORY_SMBIOS_MEMDEV + 2;
const DWORD PXS_SMBIOS_MEM_DEV_TOT_WIDTH_BIT= PXS_CATEGORY_SMBIOS_MEMDEV + 3;
const DWORD PXS_SMBIOS_MEM_DEV_DAT_WIDTH_BIT= PXS_CATEGORY_SMBIOS_MEMDEV + 4;
const DWORD PXS_SMBIOS_MEM_DEV_SIZE         = PXS_CATEGORY_SMBIOS_MEMDEV + 5;
const DWORD PXS_SMBIOS_MEM_DEV_FORM_FACTOR  = PXS_CATEGORY_SMBIOS_MEMDEV + 6;
const DWORD PXS_SMBIOS_MEM_DEV_DEVICE_SET   = PXS_CATEGORY_SMBIOS_MEMDEV + 7;
const DWORD PXS_SMBIOS_MEM_DEV_DEVICE_LOCAT = PXS_CATEGORY_SMBIOS_MEMDEV + 8;
const DWORD PXS_SMBIOS_MEM_DEV_BANK_LOCATOR = PXS_CATEGORY_SMBIOS_MEMDEV + 9;
const DWORD PXS_SMBIOS_MEM_DEV_MEMORY_TYPE  = PXS_CATEGORY_SMBIOS_MEMDEV + 10;
const DWORD PXS_SMBIOS_MEM_DEV_TYPE_DETAIL  = PXS_CATEGORY_SMBIOS_MEMDEV + 11;
const DWORD PXS_SMBIOS_MEM_DEV_SPEED_MHZ    = PXS_CATEGORY_SMBIOS_MEMDEV + 12;
const DWORD PXS_SMBIOS_MEM_DEV_MANUFACTURER = PXS_CATEGORY_SMBIOS_MEMDEV + 13;
const DWORD PXS_SMBIOS_MEM_DEV_SERIAL_NUMBER= PXS_CATEGORY_SMBIOS_MEMDEV + 14;
const DWORD PXS_SMBIOS_MEM_DEV_ASSET_TAG    = PXS_CATEGORY_SMBIOS_MEMDEV + 15;
const DWORD PXS_SMBIOS_MEM_DEV_PART_NUMBER  = PXS_CATEGORY_SMBIOS_MEMDEV + 16;
const DWORD PXS_SMBIOS_MEM_DEV_ATTRIBUTES   = PXS_CATEGORY_SMBIOS_MEMDEV + 17;
const DWORD PXS_SMBIOS_MEM_DEV_EXT_SIZE_GB  = PXS_CATEGORY_SMBIOS_MEMDEV + 18;
const DWORD PXS_SMBIOS_MEM_DEV_CLOCK_MHZ    = PXS_CATEGORY_SMBIOS_MEMDEV + 19;
const DWORD PXS_SMBIOS_MEM_DEV_MIN_VOLTAGE  = PXS_CATEGORY_SMBIOS_MEMDEV + 20;
const DWORD PXS_SMBIOS_MEM_DEV_MAX_VOLTAGE  = PXS_CATEGORY_SMBIOS_MEMDEV + 21;
const DWORD PXS_SMBIOS_MEM_DEV_CONF_VOLTAGE = PXS_CATEGORY_SMBIOS_MEMDEV + 22;

// Network Time Protocol
const DWORD PXS_CATEGORY_NET_TIME_PROTOCOL = 7200;
const DWORD PXS_NET_TIME_PROTOCOL_NAME     = PXS_CATEGORY_NET_TIME_PROTOCOL + 1;
const DWORD PXS_NET_TIME_PROTOCOL_SETTING  = PXS_CATEGORY_NET_TIME_PROTOCOL + 2;

// User Rights Assignment
const DWORD PXS_CATEGORY_USER_RIGHTS        = 7300;
const DWORD PXS_USER_RIGHTS_POLICY          = PXS_CATEGORY_USER_RIGHTS + 1;
const DWORD PXS_USER_RIGHTS_SETTING         = PXS_CATEGORY_USER_RIGHTS + 2;

// Kerberos Policy
const DWORD PXS_CATEGORY_KERBEROS_POLICY    = 7400;
const DWORD PXS_KERBEROS_POLICY_POLICY      = PXS_CATEGORY_KERBEROS_POLICY + 1;
const DWORD PXS_KERBEROS_POLICY_SETTING     = PXS_CATEGORY_KERBEROS_POLICY + 2;

// Kerberos Tickets
const DWORD PXS_CATEGORY_KERBEROS_TICKETS   = 7500;
const DWORD PXS_KERBEROS_TICKS_SERVER_NAME  = PXS_CATEGORY_KERBEROS_TICKETS + 1;
const DWORD PXS_KERBEROS_TICKS_REALM_NAME   = PXS_CATEGORY_KERBEROS_TICKETS + 2;
const DWORD PXS_KERBEROS_TICKS_START_TIME   = PXS_CATEGORY_KERBEROS_TICKETS + 3;
const DWORD PXS_KERBEROS_TICKS_END_TIME     = PXS_CATEGORY_KERBEROS_TICKETS + 4;
const DWORD PXS_KERBEROS_TICKS_RENEW_TIME   = PXS_CATEGORY_KERBEROS_TICKETS + 5;
const DWORD PXS_KERBEROS_TICKS_ENCRYPT_TYPE = PXS_CATEGORY_KERBEROS_TICKETS + 6;
const DWORD PXS_KERBEROS_TICKS_TICKET_FLAGS = PXS_CATEGORY_KERBEROS_TICKETS + 7;

// Security Registry Values
const DWORD PXS_CATEGORY_REG_SEC_VALUES     = 7600;
const DWORD PXS_REG_SEC_VALS_SUBKEY_SUBKEY  = PXS_CATEGORY_REG_SEC_VALUES + 1;
const DWORD PXS_REG_SEC_VALS_SUBKEY_SETTING = PXS_CATEGORY_REG_SEC_VALUES + 2;

// OLE DB Providers
const DWORD PXS_CATEGORY_OLE_DB_PROVIDERS   = 7700;
const DWORD PXS_OLE_DB_PROVIDERS_NAME       = PXS_CATEGORY_OLE_DB_PROVIDERS + 1;
const DWORD PXS_OLE_DB_PROVIDERS_DESCRIPTION= PXS_CATEGORY_OLE_DB_PROVIDERS + 2;

// Routing Table
const DWORD PXS_CATEGORY_ROUTING_TABLE      = 7800;
const DWORD PXS_ROUTING_TABLE_DESTINATION   = PXS_CATEGORY_ROUTING_TABLE + 1;
const DWORD PXS_ROUTING_TABLE_NETMASK       = PXS_CATEGORY_ROUTING_TABLE + 2;
const DWORD PXS_ROUTING_TABLE_NEXT_HOP      = PXS_CATEGORY_ROUTING_TABLE + 3;
const DWORD PXS_ROUTING_TABLE_INTERFACE     = PXS_CATEGORY_ROUTING_TABLE + 4;
const DWORD PXS_ROUTING_TABLE_ROUTE_TYPE    = PXS_CATEGORY_ROUTING_TABLE + 5;
const DWORD PXS_ROUTING_TABLE_PROTOCOL      = PXS_CATEGORY_ROUTING_TABLE + 6;
const DWORD PXS_ROUTING_TABLE_AGE           = PXS_CATEGORY_ROUTING_TABLE + 7;
const DWORD PXS_ROUTING_TABLE_METIC         = PXS_CATEGORY_ROUTING_TABLE + 8;

///////////////////////////////////////////////////////////////////////////////////////////////////
// Structures
///////////////////////////////////////////////////////////////////////////////////////////////////

// Hold an audit data item's idenifier and its name.
typedef struct _PXS_TYPE_PORT_SERVICE
{
    DWORD   portID;
    LPCWSTR pszServiceName;
} PXS_TYPE_PORT_SERVICE;

// Structure to for the SMBIOS SPECFICATION
typedef struct  _PXS_TYPE_SMBIOS_SPECIFICATION
{
    DWORD   itemID;
    DWORD   dataType;
    BYTE    majorVersion;
    BYTE    minorVersion;
    BYTE    smbiosType;
    BYTE    offset;
    BYTE    length;
    BYTE    bitNumber;
    wchar_t szName[ 32 ];
} PXS_TYPE_SMBIOS_SPECIFICATION;

//      ID                                Data_Type                   Maj Min Type Off  Len Bit    Name
const PXS_TYPE_SMBIOS_SPECIFICATION SMBIOS_SPECIFICATION[] =
    { { PXS_SMBIOS_BIOS_SMBIOS_VENDOR   , PXS_SMBIOS_DATA_TYPE_STRING , 2, 0,  0, 0x04,  1,  0, L"BIOS Vendor"               },
      { PXS_SMBIOS_BIOS_VERSION         , PXS_SMBIOS_DATA_TYPE_STRING , 2, 0,  0, 0x05,  1,  0, L"BIOS Version"              },
      { PXS_SMBIOS_BIOS_START_ADDRESS   , PXS_SMBIOS_DATA_TYPE_UINT16 , 2, 0,  0, 0x06,  2,  0, L"Start Address"             },
      { PXS_SMBIOS_BIOS_RELEASE_DATE    , PXS_SMBIOS_DATA_TYPE_STRING , 2, 0,  0, 0x08,  1,  0, L"Release Date"              },
      { PXS_SMBIOS_BIOS_ROM_SIZE_KB     , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 0,  0, 0x09,  1,  0, L"ROM Size"                  },
      { PXS_SMBIOS_BIOS_CHARACTERISTICS , PXS_SMBIOS_DATA_TYPE_UINT64 , 2, 0,  0, 0x0A,  8,  0, L"BIOS Characteristics"      },
      { PXS_SMBIOS_BIOS_C1_SUPPORTED    , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 0,  0, 0x0A,  8,  3, L"BIOS Charac. not supported"},
      { PXS_SMBIOS_BIOS_C1_ISA          , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 0,  0, 0x0A,  8,  4, L"ISA Support"               },
      { PXS_SMBIOS_BIOS_C1_MCA          , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 0,  0, 0x0A,  8,  5, L"MCA Support"               },
      { PXS_SMBIOS_BIOS_C1_EISA         , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 0,  0, 0x0A,  8,  6, L"EISA Support"              },
      { PXS_SMBIOS_BIOS_C1_PCI          , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 0,  0, 0x0A,  8,  7, L"PCI Support"               },
      { PXS_SMBIOS_BIOS_C1_PCMCIA       , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 0,  0, 0x0A,  8,  8, L"PCMCIA Support"            },
      { PXS_SMBIOS_BIOS_C1_PNP          , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 0,  0, 0x0A,  8,  9, L"PnP Support"               },
      { PXS_SMBIOS_BIOS_C1_APM          , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 0,  0, 0x0A,  8, 10, L"APM Support"               },
      { PXS_SMBIOS_BIOS_C1_FLASH        , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 0,  0, 0x0A,  8, 11, L"BIOS Upgradeable"          },
      { PXS_SMBIOS_BIOS_C1_SHADOW       , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 0,  0, 0x0A,  8, 12, L"BIOS Shadowing"            },
      { PXS_SMBIOS_BIOS_C1_VL_VESA      , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 0,  0, 0x0A,  8, 13, L"VL-VESA Support"           },
      { PXS_SMBIOS_BIOS_C1_SUPPORT_ESCD , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 0,  0, 0x0A,  8, 14, L"ESCD Support"              },
      { PXS_SMBIOS_BIOS_C1_BOOT_CD      , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 0,  0, 0x0A,  8, 15, L"CD Boot"                   },
      { PXS_SMBIOS_BIOS_C1_SELECT_BOOT  , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 0,  0, 0x0A,  8, 16, L"Selectable Boot"           },
      { PXS_SMBIOS_BIOS_C1_ROM_SOCKETED , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 0,  0, 0x0A,  8, 17, L"ROM Socketed"              },
      { PXS_SMBIOS_BIOS_C1_BOOT_PCMCIA  , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 0,  0, 0x0A,  8, 18, L"PCMCIA Boot"               },
      { PXS_SMBIOS_BIOS_C1_EDD          , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 0,  0, 0x0A,  8, 19, L"EDD Support"               },
      { PXS_SMBIOS_BIOS_C1_FDD_NEC_1_2MB, PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 0,  0, 0x0A,  8, 20, L"Floppy NEC 1.2MB"          },
      { PXS_SMBIOS_BIOS_C1_FDD_TOS_1_2MB, PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 0,  0, 0x0A,  8, 21, L"Floppy Toshiba 1.2MB"      },
      { PXS_SMBIOS_BIOS_C1_FDD_360KB    , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 0,  0, 0x0A,  8, 22, L"Floppy 360 KB"             },
      { PXS_SMBIOS_BIOS_C1_FDD_1_2MB    , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 0,  0, 0x0A,  8, 23, L"Floppy 1.2MB"              },
      { PXS_SMBIOS_BIOS_C1_FDD_720KB    , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 0,  0, 0x0A,  8, 24, L"Floppy 720KB"              },
      { PXS_SMBIOS_BIOS_C1_FDD_2_88MB   , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 0,  0, 0x0A,  8, 25, L"Floppy 2.88MB"             },
      { PXS_SMBIOS_BIOS_C1_PRINT_SCREEN , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 0,  0, 0x0A,  8, 26, L"Print Screen"              },
      { PXS_SMBIOS_BIOS_C1_KEYBOARD_8042, PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 0,  0, 0x0A,  8, 27, L"Keyboard 8042"             },
      { PXS_SMBIOS_BIOS_C1_SERIAL       , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 0,  0, 0x0A,  8, 28, L"Serial Services"           },
      { PXS_SMBIOS_BIOS_C1_PRINTER      , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 0,  0, 0x0A,  8, 29, L"Printer Services"          },
      { PXS_SMBIOS_BIOS_C1_CGA_MONO     , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 0,  0, 0x0A,  8, 30, L"CGA-Mono Video"            },
      { PXS_SMBIOS_BIOS_C1_NEC_PC98     , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 0,  0, 0x0A,  8, 31, L"NEC PC-98"                 },
      { PXS_SMBIOS_BIOS_CHAR_EXT_BYTE_1 , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 4,  0, 0x12,  1,  0, L"Characteristics Ext. 1"    },
      { PXS_SMBIOS_BIOS_CX1_SUPPORT_ACPI, PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 4,  0, 0x12,  1,  0, L"ACPI Support"              },
      { PXS_SMBIOS_BIOS_CX1_LEGACY_USB  , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 4,  0, 0x12,  1,  1, L"USB Support"               },
      { PXS_SMBIOS_BIOS_CX1_SUPPORT_AGP , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 4,  0, 0x12,  1,  2, L"AGP Support"               },
      { PXS_SMBIOS_BIOS_CX1_BOOT_I2O    , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 4,  0, 0x12,  1,  3, L"I2O Boot"                  },
      { PXS_SMBIOS_BIOS_CX1_BOOT_LS120  , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 4,  0, 0x12,  1,  4, L"LS-120 Boot"               },
      { PXS_SMBIOS_BIOS_CX1_BOOT_ATAPI_Z, PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 4,  0, 0x12,  1,  5, L"ATAPI ZIP Boot"            },
      { PXS_SMBIOS_BIOS_CX1_BOOT_1394   , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 4,  0, 0x12,  1,  6, L"Firewire Boot"             },
      { PXS_SMBIOS_BIOS_CX1_SMART_BAT   , PXS_SMBIOS_DATA_TYPE_BOOLEAN, 2, 4,  0, 0x12,  1,  7, L"Smart Battery"             },
      { PXS_SMBIOS_BIOS_CHAR_EXT_BYTE_2 , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 4,  0, 0x13,  1,  0, L"Characteristics Ext. 2"    },
      { PXS_SMBIOS_BIOS_SYS_BIOS_MAJOR  , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 4,  0, 0x14,  1,  0, L"System BIOS Major"         },
      { PXS_SMBIOS_BIOS_SYS_BIOS_MINOR  , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 4,  0, 0x15,  1,  0, L"System BIOS Minor"         },
      { PXS_SMBIOS_BIOS_FIRMWARE_MAJOR  , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 4,  0, 0x16,  1,  0, L"Firmware Major"            },
      { PXS_SMBIOS_BIOS_FIRMWARE_MINOR  , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 4,  0, 0x17,  1,  0, L"Firmware Minor"            },
      { PXS_SMBIOS_SYSINFO_MANUFACTURER , PXS_SMBIOS_DATA_TYPE_STRING , 2, 0,  1, 0x04,  1,  0, L"Manufacturer"              },
      { PXS_SMBIOS_SYSINFO_PRODUCT_NAME , PXS_SMBIOS_DATA_TYPE_STRING , 2, 0,  1, 0x05,  1,  0, L"Product Name"              },
      { PXS_SMBIOS_SYSINFO_VERSION      , PXS_SMBIOS_DATA_TYPE_STRING , 2, 0,  1, 0x06,  1,  0, L"Version"                   },
      { PXS_SMBIOS_SYSINFO_SERIAL_NUMBER, PXS_SMBIOS_DATA_TYPE_STRING , 2, 0,  1, 0x07,  1,  0, L"Serial Number"             },
      { PXS_SMBIOS_SYSINFO_UUID         , PXS_SMBIOS_DATA_TYPE_STRING , 2, 1,  1, 0x08, 16,  0, L"Universal Unique ID"       },
      { PXS_SMBIOS_SYSINFO_WAKEUP_TYPE  , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 1,  1, 0x18,  1,  0, L"Wake-Up Type"              },
      { PXS_SMBIOS_SYSINFO_SKU_NUMBER   , PXS_SMBIOS_DATA_TYPE_STRING , 2, 4,  1, 0x19,  1,  0, L"SKU Number"                },
      { PXS_SMBIOS_SYSINFO_FAMILY       , PXS_SMBIOS_DATA_TYPE_STRING , 2, 4,  1, 0x1A,  1,  0, L"Product Family"            },
      { PXS_SMBIOS_BOARD_MANUFACTURER   , PXS_SMBIOS_DATA_TYPE_STRING , 2, 0,  2, 0x04,  1,  0, L"Manufacturer"              },
      { PXS_SMBIOS_BOARD_PRODUCT        , PXS_SMBIOS_DATA_TYPE_STRING , 2, 0,  2, 0x05,  1,  0, L"Product"                   },
      { PXS_SMBIOS_BOARD_VERSION        , PXS_SMBIOS_DATA_TYPE_STRING , 2, 0,  2, 0x06,  1,  0, L"Version"                   },
      { PXS_SMBIOS_BOARD_SERIAL_NUMBER  , PXS_SMBIOS_DATA_TYPE_STRING , 2, 0,  2, 0x07,  1,  0, L"Serial Number"             },
      { PXS_SMBIOS_BOARD_ASSET_TAG      , PXS_SMBIOS_DATA_TYPE_STRING , 2, 0,  2, 0x08,  1,  0, L"Asset Tag"                 },
      { PXS_SMBIOS_BOARD_FEATURES_FLAG  , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 0,  2, 0x09,  1,  0, L"Features"                  },
      { PXS_SMBIOS_BOARD_BOARD_TYPE     , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 0,  2, 0x0D,  1,  0, L"Board Type"                },
      { PXS_SMBIOS_CHASSIS_MANUFACTURER , PXS_SMBIOS_DATA_TYPE_STRING , 2, 0,  3, 0x04,  1,  0, L"Manufacturer"              },
      { PXS_SMBIOS_CHASSIS_TYPE         , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 0,  3, 0x05,  1,  0, L"Chassis Type"              },
      { PXS_SMBIOS_CHASSIS_VERSION      , PXS_SMBIOS_DATA_TYPE_STRING , 2, 0,  3, 0x06,  1,  0, L"Version"                   },
      { PXS_SMBIOS_CHASSIS_SERIAL_NUMBER, PXS_SMBIOS_DATA_TYPE_STRING , 2, 0,  3, 0x07,  1,  0, L"Serial Number"             },
      { PXS_SMBIOS_CHASSIS_ASSET_TAG_NUM, PXS_SMBIOS_DATA_TYPE_STRING , 2, 0,  3, 0x08,  1,  0, L"Asset Tag"                 },
      { PXS_SMBIOS_CHASSIS_BOOT_UP_STATE, PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 1,  3, 0x09,  1,  0, L"Boot Up State"             },
      { PXS_SMBIOS_CHASSIS_POWER_SUPPLY , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 1,  3, 0x0A,  1,  0, L"Power State"               },
      { PXS_SMBIOS_CHASSIS_THERMAL_STATE, PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 1,  3, 0x0B,  1,  0, L"Thermal State"             },
      { PXS_SMBIOS_CHASSIS_SECUR_STATUS , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 1,  3, 0x0C,  1,  0, L"Security Status"           },
      { PXS_SMBIOS_CHASSIS_OEM_DEFINED  , PXS_SMBIOS_DATA_TYPE_UINT32 , 2, 3,  3, 0x0D,  4,  0, L"OEM Defined"               },
      { PXS_SMBIOS_CHASSIS_HEIGHT_U     , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 3,  3, 0x11,  1,  0, L"Height"                    },
      { PXS_SMBIOS_CHASSIS_POWER_CORDS  , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 3,  3, 0x12,  1,  0, L"Number Of Power Cords"     },
      { PXS_SMBIOS_PROCESSOR_SOCKET     , PXS_SMBIOS_DATA_TYPE_STRING , 2, 0,  4, 0x04,  1,  0, L"Socket Designation"        },
      { PXS_SMBIOS_PROCESSOR_TYPE       , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 0,  4, 0x05,  1,  0, L"Processor Type"            },
      { PXS_SMBIOS_PROCESSOR_FAMILY     , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 0,  4, 0x06,  1,  0, L"Processor Family"          },
      { PXS_SMBIOS_PROCESSOR_MANUFAC    , PXS_SMBIOS_DATA_TYPE_STRING , 2, 0,  4, 0x07,  1,  0, L"Processor Manufacturer"    },
      { PXS_SMBIOS_PROCESSOR_ID         , PXS_SMBIOS_DATA_TYPE_UINT64 , 2, 0,  4, 0x08,  8,  0, L"Processor ID"              },
      { PXS_SMBIOS_PROCESSOR_VERSION    , PXS_SMBIOS_DATA_TYPE_STRING , 2, 0,  4, 0x10,  1,  0, L"Processor Version"         },
      { PXS_SMBIOS_PROCESSOR_VOLTAGE    , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 0,  4, 0x11,  1,  0, L"Voltage"                   },
      { PXS_SMBIOS_PROCESSOR_EXT_CLK_MHZ, PXS_SMBIOS_DATA_TYPE_UINT16 , 2, 0,  4, 0x12,  2,  0, L"External Clock"            },
      { PXS_SMBIOS_PROCESSOR_MAXIMUM_MHZ, PXS_SMBIOS_DATA_TYPE_UINT16 , 2, 0,  4, 0x14,  2,  0, L"Maximum Speed"             },
      { PXS_SMBIOS_PROCESSOR_CURRENT_MHZ, PXS_SMBIOS_DATA_TYPE_UINT16 , 2, 0,  4, 0x16,  2,  0, L"Current Speed"             },
      { PXS_SMBIOS_PROCESSOR_STATUS     , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 0,  4, 0x18,  1,  0, L"Status"                    },
      { PXS_SMBIOS_PROCESSOR_UPGRADE    , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 0,  4, 0x19,  1,  0, L"Processor Upgrade"         },
      { PXS_SMBIOS_PROCESSOR_SERIAL_NUM , PXS_SMBIOS_DATA_TYPE_STRING , 2, 3,  4, 0x20,  1,  0, L"Serial Number"             },
      { PXS_SMBIOS_PROCESSOR_ASSET_TAG  , PXS_SMBIOS_DATA_TYPE_STRING , 2, 3,  4, 0x21,  1,  0, L"Asset Tag"                 },
      { PXS_SMBIOS_PROCESSOR_PART_NUM   , PXS_SMBIOS_DATA_TYPE_STRING , 2, 3,  4, 0x22,  1,  0, L"Part Number"               },
      { PXS_SMBIOS_PROCESSOR_CORE_COUNT , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 5,  4, 0x23,  1,  0, L"Core Count"                },
      { PXS_SMBIOS_PROCESSOR_CORE_ENABLE, PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 5,  4, 0x24,  1,  0, L"Core Enabled"              },
      { PXS_SMBIOS_PROCESSOR_THREADS    , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 5,  4, 0x25,  1,  0, L"Thread Count"              },
      { PXS_SMBIOS_PROCESSOR_CHARACTERIS, PXS_SMBIOS_DATA_TYPE_UINT16 , 2, 5,  4, 0x26,  2,  0, L"Processor Characteristics" },
      { PXS_SMBIOS_PROCESSOR_FAMILY_2   , PXS_SMBIOS_DATA_TYPE_UINT16 , 2, 6,  4, 0x28,  1,  0, L"Processor Family 2"        },
      { PXS_SMBIOS_MEMCTRL_ERR_DETECT   , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 0,  5, 0x04,  1,  0, L"Error Detecting"           },
      { PXS_SMBIOS_MEMCTRL_ERR_CORRECT  , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 0,  5, 0x05,  1,  0, L"Error Correcting"          },
      { PXS_SMBIOS_MEMCTRL_SUPPORT_INTER, PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 0,  5, 0x06,  1,  0, L"Supported Interleave"      },
      { PXS_SMBIOS_MEMCTRL_CURRENT_INTER, PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 0,  5, 0x07,  1,  0, L"Current Interleave"        },
      { PXS_SMBIOS_MEMCTRL_MAX_MODULE_MB, PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 0,  5, 0x08,  1,  0, L"Max. Memory Module"        },
      { PXS_SMBIOS_MEMCTRL_SPEEDS       , PXS_SMBIOS_DATA_TYPE_UINT16 , 2, 0,  5, 0x09,  2,  0, L"Supported Memory Speeds"   },
      { PXS_SMBIOS_MEMCTRL_MEM_TYPES    , PXS_SMBIOS_DATA_TYPE_UINT16 , 2, 0,  5, 0x0B,  2,  0, L"Supported Memory Types"    },
      { PXS_SMBIOS_MEMCTRL_VOLTAGE      , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 0,  5, 0x0D,  1,  0, L"Memory Module Voltage"     },
      { PXS_SMBIOS_MEMCTRL_NUM_SLOTS    , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 0,  5, 0x0E,  1,  0, L"Number Memory Slots"       },
      { PXS_SMBIOS_MEMMODULE_SOCKET     , PXS_SMBIOS_DATA_TYPE_STRING , 2, 0,  6, 0x04,  1,  0, L"Socket Designation"        },
      { PXS_SMBIOS_MEMMODULE_BANK_CONN  , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 0,  6, 0x05,  1,  0, L"Bank Connection"           },
      { PXS_SMBIOS_MEMMODULE_SPEED_NS   , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 0,  6, 0x06,  1,  0, L"Current Speed"             },
      { PXS_SMBIOS_MEMMODULE_MEM_TYPE   , PXS_SMBIOS_DATA_TYPE_UINT16 , 2, 0,  6, 0x07,  2,  0, L"Current Memory Type"       },
      { PXS_SMBIOS_MEMMODULE_INSTAL_SIZE, PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 0,  6, 0x09,  1,  0, L"Installed Size"            },
      { PXS_SMBIOS_MEMMODULE_ENABLE_SIZE, PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 0,  6, 0x0A,  1,  0, L"Enabled Size"              },
      { PXS_SMBIOS_MEMMODULE_ERR_STATUS , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 0,  6, 0x0B,  1,  0, L"Error Status"              },
      { PXS_SMBIOS_CPUCACHE_SOCKET      , PXS_SMBIOS_DATA_TYPE_STRING , 2, 0,  7, 0x04,  1,  0, L"Socket Designation"        },
      { PXS_SMBIOS_CPUCACHE_CONFIG      , PXS_SMBIOS_DATA_TYPE_UINT16 , 2, 0,  7, 0x05,  2,  0, L"Cache Configuration"       },
      { PXS_SMBIOS_CPUCACHE_MAX_SIZE_KB , PXS_SMBIOS_DATA_TYPE_UINT16 , 2, 0,  7, 0x07,  2,  0, L"Max. Cache Size"           },
      { PXS_SMBIOS_CPUCACHE_INST_SIZE_KB, PXS_SMBIOS_DATA_TYPE_UINT16 , 2, 0,  7, 0x09,  2,  0, L"Installed Cache"           },
      { PXS_SMBIOS_CPUCACHE_SUPPORT_SRAM, PXS_SMBIOS_DATA_TYPE_UINT16 , 2, 0,  7, 0x0B,  2,  0, L"Supported SRAM Type"       },
      { PXS_SMBIOS_CPUCACHE_CURRENT_SRAM, PXS_SMBIOS_DATA_TYPE_UINT16 , 2, 0,  7, 0x0D,  2,  0, L"Current SRAM Type"         },
      { PXS_SMBIOS_CPUCACHE_SPEED_NS    , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 1,  7, 0x0F,  1,  0, L"Cache Speed"               },
      { PXS_SMBIOS_CPUCACHE_ERR_CORRECT , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 1,  7, 0x10,  1,  0, L"Error Correction Type"     },
      { PXS_SMBIOS_CPUCACHE_SYSTEM_TYPE , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 1,  7, 0x11,  1,  0, L"System Cache Type"         },
      { PXS_SMBIOS_CPUCACHE_ASSOCIATVITY, PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 1,  7, 0x12,  1,  0, L"Associativity"             },
      { PXS_SMBIOS_PORTCONN_INT_REF     , PXS_SMBIOS_DATA_TYPE_STRING , 2, 0,  8, 0x04,  1,  0, L"Internal Designation"      },
      { PXS_SMBIOS_PORTCONN_INT_CONN    , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 0,  8, 0x05,  1,  0, L"Internal Connection Type"  },
      { PXS_SMBIOS_PORTCONN_EXT_REF     , PXS_SMBIOS_DATA_TYPE_STRING , 2, 0,  8, 0x06,  1,  0, L"External Designation"      },
      { PXS_SMBIOS_PORTCONN_EXT_CONN    , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 0,  8, 0x07,  1,  0, L"External Connection Type"  },
      { PXS_SMBIOS_PORTCONN_PORT_TYPE   , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 0,  8, 0x08,  1,  0, L"Port Type"                 },
      { PXS_SMBIOS_SYSSLOT_DESIGNATION  , PXS_SMBIOS_DATA_TYPE_STRING , 2, 0,  9, 0x04,  1,  0, L"Slot Designation"          },
      { PXS_SMBIOS_SYSSLOT_SLOT_TYPE    , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 0,  9, 0x05,  1,  0, L"Slot Type"                 },
      { PXS_SMBIOS_SYSSLOT_BUS_WIDTH    , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 0,  9, 0x06,  1,  0, L"Data Bus Width"            },
      { PXS_SMBIOS_SYSSLOT_CURRENT_USAGE, PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 0,  9, 0x07,  1,  0, L"Current Usage"             },
      { PXS_SMBIOS_SYSSLOT_SLOT_LENGTH  , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 0,  9, 0x08,  1,  0, L"Slot Length"               },
      { PXS_SMBIOS_SYSSLOT_SLOT_ID      , PXS_SMBIOS_DATA_TYPE_UINT16 , 2, 0,  9, 0x09,  2,  0, L"Slot ID"                   },
      { PXS_SMBIOS_SYSSLOT_C_1          , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 0,  9, 0x0B,  1,  0, L"Slot Characteristics 1"    },
      { PXS_SMBIOS_SYSSLOT_C_2          , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 1,  9, 0x0C,  1,  0, L"Slot Characteristics 2"    },
      { PXS_SMBIOS_SYSSLOT_SEGMENT_NUM  , PXS_SMBIOS_DATA_TYPE_UINT16 , 2, 6,  9, 0x0D,  2,  0, L"Segment Group Number"      },
      { PXS_SMBIOS_SYSSLOT_BUS_NUM      , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 6,  9, 0x0F,  1,  0, L"Bus Number"                },
      { PXS_SMBIOS_SYSSLOT_DEVICE_NUM   , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 6,  9, 0x10,  1,  0, L"Device/Function Number"    },
      { PXS_SMBIOS_MEMARRAY_LOCATION    , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 1, 16, 0x04,  1,  0, L"Location"                  },
      { PXS_SMBIOS_MEMARRAY_USE         , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 1, 16, 0x05,  1,  0, L"Use Item"                  },
      { PXS_SMBIOS_MEMARRAY_ERR_CORRECT , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 1, 16, 0x06,  1,  0, L"Error Correction"          },
      { PXS_SMBIOS_MEMARRAY_MAX_CAP_KB  , PXS_SMBIOS_DATA_TYPE_UINT32 , 2, 1, 16, 0x07,  4,  0, L"Maximum Capacity"          },
      { PXS_SMBIOS_MEMARRAY_ERROR_HANDLE, PXS_SMBIOS_DATA_TYPE_UINT16 , 2, 1, 16, 0x0B,  2,  0, L"Error Handle"              },
      { PXS_SMBIOS_MEMARRAY_NUM_DEVICES , PXS_SMBIOS_DATA_TYPE_UINT16 , 2, 1, 16, 0x0D,  2,  0, L"Number of Devices"         },
      { PXS_SMBIOS_MEMARRAY_EX_MAX_CAP  , PXS_SMBIOS_DATA_TYPE_UINT64 , 2, 7, 16, 0x0F,  8,  0, L"Extended Maximum Capacity" },
      { PXS_SMBIOS_MEM_DEV_ERROR_HANDLE , PXS_SMBIOS_DATA_TYPE_UINT16 , 2, 1, 17, 0x06,  2,  0, L"Error Handle"              },
      { PXS_SMBIOS_MEM_DEV_TOT_WIDTH_BIT, PXS_SMBIOS_DATA_TYPE_UINT16 , 2, 1, 17, 0x08,  2,  0, L"Total Width"               },
      { PXS_SMBIOS_MEM_DEV_DAT_WIDTH_BIT, PXS_SMBIOS_DATA_TYPE_UINT16 , 2, 1, 17, 0x0A,  2,  0, L"Data Width"                },
      { PXS_SMBIOS_MEM_DEV_SIZE         , PXS_SMBIOS_DATA_TYPE_UINT16 , 2, 1, 17, 0x0C,  2,  0, L"Device Size"               },
      { PXS_SMBIOS_MEM_DEV_FORM_FACTOR  , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 1, 17, 0x0E,  1,  0, L"Form Factor"               },
      { PXS_SMBIOS_MEM_DEV_DEVICE_SET   , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 1, 17, 0x0F,  1,  0, L"Device Set"                },
      { PXS_SMBIOS_MEM_DEV_DEVICE_LOCAT , PXS_SMBIOS_DATA_TYPE_STRING , 2, 1, 17, 0x10,  1,  0, L"Device Locator"            },
      { PXS_SMBIOS_MEM_DEV_BANK_LOCATOR , PXS_SMBIOS_DATA_TYPE_STRING , 2, 1, 17, 0x11,  1,  0, L"Bank Locator"              },
      { PXS_SMBIOS_MEM_DEV_MEMORY_TYPE  , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 1, 17, 0x12,  1,  0, L"Memory Type"               },
      { PXS_SMBIOS_MEM_DEV_TYPE_DETAIL  , PXS_SMBIOS_DATA_TYPE_UINT16 , 2, 1, 17, 0x13,  2,  0, L"Memory Type Detail"        },
      { PXS_SMBIOS_MEM_DEV_SPEED_MHZ    , PXS_SMBIOS_DATA_TYPE_UINT16 , 2, 3, 17, 0x15,  2,  0, L"Speed"                     },
      { PXS_SMBIOS_MEM_DEV_MANUFACTURER , PXS_SMBIOS_DATA_TYPE_STRING , 2, 3, 17, 0x17,  1,  0, L"Manufacturer"              },
      { PXS_SMBIOS_MEM_DEV_SERIAL_NUMBER, PXS_SMBIOS_DATA_TYPE_STRING , 2, 3, 17, 0x18,  1,  0, L"Serial Number"             },
      { PXS_SMBIOS_MEM_DEV_ASSET_TAG    , PXS_SMBIOS_DATA_TYPE_STRING , 2, 3, 17, 0x19,  1,  0, L"Asset Tag"                 },
      { PXS_SMBIOS_MEM_DEV_PART_NUMBER  , PXS_SMBIOS_DATA_TYPE_STRING , 2, 3, 17, 0x1A,  1,  0, L"Part Number"               },
      { PXS_SMBIOS_MEM_DEV_ATTRIBUTES   , PXS_SMBIOS_DATA_TYPE_UINT8  , 2, 7, 17, 0x1B,  1,  0, L"Attributes"                },
      { PXS_SMBIOS_MEM_DEV_EXT_SIZE_GB  , PXS_SMBIOS_DATA_TYPE_UINT32 , 2, 7, 17, 0x1C,  4,  0, L"Extended Size"             },
      { PXS_SMBIOS_MEM_DEV_CLOCK_MHZ    , PXS_SMBIOS_DATA_TYPE_UINT16 , 2, 7, 17, 0x20,  2,  0, L"Config. Memory Clock Speed"},
      { PXS_SMBIOS_MEM_DEV_MIN_VOLTAGE  , PXS_SMBIOS_DATA_TYPE_UINT16 , 2, 8, 17, 0x22,  2,  0, L"Minimum voltage"           },
      { PXS_SMBIOS_MEM_DEV_MAX_VOLTAGE  , PXS_SMBIOS_DATA_TYPE_UINT16 , 2, 8, 17, 0x24,  2,  0, L"Maximum voltage"           },
      { PXS_SMBIOS_MEM_DEV_CONF_VOLTAGE , PXS_SMBIOS_DATA_TYPE_UINT16 , 2, 8, 17, 0x26,  2,  0, L"Configured voltage"        },
    };

// Structure to hold the properties for reporting audit data
typedef struct _TYPE_DATA_CATEGORY_PROPERTIES
{
    DWORD   categoryID;
    DWORD   captionID;
    DWORD   isNode;
    BYTE    depth;
    DWORD   isColumnar;
    DWORD   nameStringID;
} TYPE_DATA_CATEGORY_PROPERTIES;

// Initialise an array of reporting structures
const TYPE_DATA_CATEGORY_PROPERTIES PXS_DATA_CATEGORY_PROPERTIES[] =
{
    // categoryID                      captionID                      isNode  dep isColumnar pszName
    { PXS_CATEGORY_INSTALLED_SOFTWARE, 0                              , TRUE , 0, FALSE, PXS_IDS_1122_INSTALLED_SOFTWARE     },
    { PXS_CATEGORY_SECURITY          , 0                              , TRUE , 0, FALSE, PXS_IDS_1125_SECURITY               },
    { PXS_CATEGORY_GROUPS_AND_USERS  , 0                              , TRUE , 0, FALSE, PXS_IDS_1126_GROUPS_AND_USERS       },
    { PXS_CATEGORY_WINDOWS_NETWORK   , 0                              , TRUE , 0, FALSE, PXS_IDS_1132_WINDOWS_NETWORK        },
    { PXS_CATEGORY_NETWORK_TCPIP     , 0                              , TRUE , 0, FALSE, PXS_IDS_1133_NETWORK_TCPIP          },
    { PXS_CATEGORY_SMBIOS            , 0                              , TRUE , 0, FALSE, PXS_IDS_1139_SYSTEM_MANAGEMENT      },
    { PXS_CATEGORY_ODBC_INFORMATION  , 0                              , TRUE , 0, FALSE, PXS_IDS_1148_ODBC_INFORMATION       },
    { PXS_CATEGORY_SYSTEM_OVERVIEW   , 0                              , FALSE, 0, TRUE , PXS_IDS_1121_SYSTEM_OVERVIEW        },
    { PXS_CATEGORY_ACTIVE_SETUP      , PXS_ACTIVE_SETUP_NAME          , FALSE, 1, FALSE, PXS_IDS_1161_ACTIVE_SETUP           },
    { PXS_CATEGORY_INSTALLED_PROGS   , PXS_INSTAL_PROGS_NAME          , TRUE , 1, TRUE , PXS_IDS_1162_INSTALLED_PROGRAMS     },
    { PXS_CATEGORY_SOFTWARE_UPDATES  , PXS_SOFTWARE_UPDATES_UPDATE_ID , FALSE, 1, FALSE, PXS_IDS_1163_SOFTWARE_UPDATES       },
    { PXS_CATEGORY_OS                , 0                              , FALSE, 0, TRUE , PXS_IDS_1123_OPERATING_SYSTEM       },
    { PXS_CATEGORY_PERIPHERALS       , PXS_PERIPHERALS_NAME           , FALSE, 0, FALSE, PXS_IDS_1124_PERIPHERALS            },
    { PXS_CATEGORY_OPEN_PORTS        , PXS_OPEN_PORTS_CAPTION         , TRUE , 1, TRUE , PXS_IDS_1167_OPEN_PORTS             },
    { PXS_CATEGORY_SECURITY_LOG      , 0                              , FALSE, 1, FALSE, PXS_IDS_1170_SECURITY_LOG           },
    { PXS_CATEGORY_SECUR_SETTINGS    , PXS_SECURITY_SETTINGS_ITEM     , FALSE, 1, FALSE, PXS_IDS_1171_SECURITY_SETTINGS      },
    { PXS_CATEGORY_SYSRESTORE        , PXS_SYSRESTORE_SEQUENCE        , FALSE, 1, FALSE, PXS_IDS_1172_SYSTEM_RESTORE         },
    { PXS_CATEGORY_USERPRIVS         , PXS_USERPRIVS_PRIVILEGE_NAME   , FALSE, 1, FALSE, PXS_IDS_1174_USER_PRIVILEGES        },
    { PXS_CATEGORY_WINDOWS_FIREWALL  , PXS_WINDOWS_FIREWALL_NAME      , FALSE, 1, FALSE, PXS_IDS_1175_WINDOWS_FIREWALL       },
    { PXS_CATEGORY_GROUPS            , PXS_GROUPS_NAME                , TRUE , 1, TRUE , PXS_IDS_1176_GROUPS                 },
    { PXS_CATEGORY_GROUPMEMBERS      , 0                              , FALSE, 1, FALSE, PXS_IDS_1177_GROUP_MEMBERS          },
    { PXS_CATEGORY_GROUPPOLICY       , 0                              , FALSE, 1, FALSE, PXS_IDS_1178_GROUP_POLICY           },
    { PXS_CATEGORY_USERS             , PXS_USERS_USERNAME             , TRUE , 1, TRUE , PXS_IDS_1179_USERS                  },
    { PXS_CATEGORY_SCHED_TASKS       , PXS_SCHED_TASKS_TASK_NAME      , TRUE , 0, TRUE , PXS_IDS_1127_SCHEDULED_TASKS        },
    { PXS_CATEGORY_UPTIME            , 0                              , FALSE, 0, TRUE , PXS_IDS_1128_UPTIME_STATISITCS      },
    { PXS_CATEGORY_ERROR_LOG         , 0                              , FALSE, 0, FALSE, PXS_IDS_1129_ERROR_LOGS             },
    { PXS_CATEGORY_WIN_NET_FILES     , PXS_WIN_NET_FILES_PATHNAME     , FALSE, 1, FALSE, PXS_IDS_1180_NETWORK_FILES          },
    { PXS_CATEGORY_WIN_NET_SESSIONS  , PXS_WIN_NET_SESS_COMPUTERNAME  , FALSE, 1, FALSE, PXS_IDS_1181_NETWORK_SESSIONS       },
    { PXS_CATEGORY_WIN_NET_SHARES    , PXS_WIN_NET_SHARES_NETNAME     , FALSE, 1, FALSE, PXS_IDS_1182_NETWORK_SHARES         },
    { PXS_CATEGORY_NET_ADAPTERS      , PXS_NET_ADAPT_ADAPTER_NAME     , TRUE , 1, TRUE , PXS_IDS_1197_NETWORK_ADAPTERS       },
    { PXS_CATEGORY_HARDWARE_DEVICES  , PXS_HARDWARE_DEVS_DEVICE_TYPE  , TRUE , 0, TRUE , PXS_IDS_1134_HARDWARE_DEVICES       },
    { PXS_CATEGORY_PRINTERS          , PXS_PRINTERS_NAME              , TRUE , 0, TRUE , PXS_IDS_1137_INSTALLED_PRINTERS     },
    { PXS_CATEGORY_BIOS_VERSION      , 0                              , FALSE, 0, TRUE , PXS_IDS_1138_BIOS_VERSION           },
    { PXS_CATEGORY_CPU_BASIC         , PXS_CPU_BASIC_FULL_NAME        , TRUE , 0, TRUE , PXS_IDS_1140_PROCESSORS             },
    { PXS_CATEGORY_MEMORY            , 0                              , FALSE, 0, TRUE , PXS_IDS_1141_MEMORY                 },
    { PXS_CATEGORY_PHYS_DISKS        , PXS_PHYS_DISKS_MODEL           , TRUE , 0, TRUE , PXS_IDS_1142_PHYSICAL_DISKS         },
    { PXS_CATEGORY_DRIVES            , PXS_DRIVES_LETTER              , TRUE , 0, TRUE , PXS_IDS_1143_DRIVES                 },
    { PXS_CATEGORY_COMMPORTS         , PXS_COMMPORTS_NAME             , TRUE , 0, TRUE , PXS_IDS_1144_COMMUNICATION_PORTS    },
    { PXS_CATEGORY_STARTUP           , PXS_STARTUP_PROGRAM_NAME       , FALSE, 0, FALSE, PXS_IDS_1145_STARTUP_PROGRAMMES     },
    { PXS_CATEGORY_NTSERVICES        , PXS_NTSERVICES_SERVICE_NAME    , TRUE , 0, TRUE , PXS_IDS_1146_SERVICES               },
    { PXS_CATEGORY_RUNNING_PROCS     , PXS_RUNNING_PROCS_NAME         , FALSE, 0, FALSE, PXS_IDS_1147_RUNNING_PROGRAMMES     },
    { PXS_CATEGORY_DISPLAY_EDID      , PXS_DISPLAY_EDID_DISPLAY_NAME  , TRUE , 0, TRUE , PXS_IDS_1135_DISPLAYS               },
    { PXS_CATEGORY_DISPLAY_ADAPTERS  , PXS_DISP_ADAPT_NAME            , TRUE , 0, TRUE , PXS_IDS_1136_DISPLAY_ADAPTERS       },
    { PXS_CATEGORY_ENVIRON_VARS      , 0                              , FALSE, 0, FALSE, PXS_IDS_1130_ENVIRONMENT_VARIABLES  },
    { PXS_CATEGORY_REGION_SETTINGS   , 0                              , FALSE, 0, FALSE, PXS_IDS_1131_REGIONAL_SETTINGS      },
    { PXS_CATEGORY_PERMISSIONS       , PXS_PERMISSIONS_OBJECT_NAME    , TRUE , 1, TRUE , PXS_IDS_1168_PERMISSIONS            },
    { PXS_CATEGORY_USER_LOGONS       , PXS_USER_LOGONS_USER_NAME      , TRUE , 0, TRUE , PXS_IDS_1151_USER_LOGON_STATISTICS  },
    { PXS_CATEGORY_SOFTWARE_METERING , PXS_SOFT_METER_FILE_NAME       , TRUE , 0, TRUE , PXS_IDS_1150_SOFTWARE_METERING      },
    { PXS_CATEGORY_ODBC_DRIVERS      , PXS_ODBC_DRIVERS_NAME          , TRUE , 1, TRUE , PXS_IDS_1196_ODBC_DRIVERS           },
    { PXS_CATEGORY_ODBC_DATA_SOURCES , 0                              , FALSE, 1, FALSE, PXS_IDS_1195_ODBC_DATA_SOURCES      },
    { PXS_CATEGORY_SMBIOS_INFO       , 0                              , FALSE, 1, TRUE , PXS_IDS_1183_BIOS_DETAILS           },
    { PXS_CATEGORY_SMBIOS_SYSINFO    , 0                              , FALSE, 1, TRUE , PXS_IDS_1184_SYSTEM_INFORMATION     },
    { PXS_CATEGORY_SMBIOS_BOARD      , PXS_SMBIOS_BOARD_PRODUCT       , TRUE , 1, TRUE , PXS_IDS_1185_BASE_BOARD             },
    { PXS_CATEGORY_SMBIOS_CHASSIS    , 0                              , FALSE, 1, TRUE , PXS_IDS_1186_CHASSIS                },
    { PXS_CATEGORY_SMBIOS_PROC       , PXS_SMBIOS_PROCESSOR_VERSION   , TRUE , 1, TRUE , PXS_IDS_1187_PROCESSOR              },
    { PXS_CATEGORY_SMBIOS_MEMCTRL    , PXS_SMBIOS_MEMCTRL_ITEM_NUMBER , TRUE , 1, TRUE , PXS_IDS_1188_MEMORY_CONTROLLER      },
    { PXS_CATEGORY_SMBIOS_MEMMODULE  , PXS_SMBIOS_MEMMODULE_SOCKET    , TRUE , 1, TRUE , PXS_IDS_1189_MEMORY_MODULE          },
    { PXS_CATEGORY_SMBIOS_CPUCACHE   , PXS_SMBIOS_CPUCACHE_SOCKET     , TRUE , 1, TRUE , PXS_IDS_1190_CACHE                  },
    { PXS_CATEGORY_SMBIOS_PORTCONN   , PXS_SMBIOS_PORTCONN_INT_REF    , TRUE , 1, TRUE , PXS_IDS_1191_PORT_CONNECTOR         },
    { PXS_CATEGORY_SMBIOS_SYSSLOT    , PXS_SMBIOS_SYSSLOT_DESIGNATION , TRUE , 1, TRUE , PXS_IDS_1192_MEMORY_SLOTS           },
    { PXS_CATEGORY_SMBIOS_MEMARRAY   , PXS_SMBIOS_MEMARRAY_ITEM_NUMBER, FALSE, 1, TRUE , PXS_IDS_1193_MEMORY_ARRAY           },
    { PXS_CATEGORY_SMBIOS_MEMDEV     , PXS_SMBIOS_MEM_DEV_ITEM_NUMBER , TRUE , 1, TRUE , PXS_IDS_1194_MEMORY_DEVICE          },
    { PXS_CATEGORY_NET_TIME_PROTOCOL , 0                              , FALSE, 1, FALSE, PXS_IDS_1166_NETWORK_TIME_PROTOCOL  },
    { PXS_CATEGORY_USER_RIGHTS       , 0                              , FALSE, 1, FALSE, PXS_IDS_1173_USER_RIGHTS_ASSIGNMENT },
    { PXS_CATEGORY_KERBEROS_POLICY   , 0                              , FALSE, 1, FALSE, PXS_IDS_1164_KERBEROS_POLICY        },
    { PXS_CATEGORY_KERBEROS_TICKETS  , PXS_KERBEROS_TICKS_SERVER_NAME , TRUE , 1, TRUE , PXS_IDS_1165_KERBEROS_TICKETS       },
    { PXS_CATEGORY_REG_SEC_VALUES    , 0                              , FALSE, 1, FALSE, PXS_IDS_1169_REGISTRY_SECURITY_VALS },
    { PXS_CATEGORY_OLE_DB_PROVIDERS  , 0                              , FALSE, 0, FALSE, PXS_IDS_1149_OLE_DB_PROVIDERS       },
    { PXS_CATEGORY_ROUTING_TABLE     , PXS_ROUTING_TABLE_DESTINATION  , TRUE , 1, TRUE , PXS_IDS_1198_ROUTING_TABLE          }
};

// Hold an audit data item's idenifier and its name.
typedef struct _PXS_TYPE_AUDIT_ITEM
{
    DWORD   itemID;
    LPCWSTR pszName;
} PXS_TYPE_AUDIT_ITEM;

// Array is ordered by itemID for use with bsearch
const PXS_TYPE_AUDIT_ITEM PXS_AUDIT_ITEMS[] = {
    { PXS_SYS_OVERVIEW_COMPUTER_NAME  , L"Computer Name"                 },
    { PXS_SYS_OVERVIEW_DOMAIN_NAME    , L"Domain Name"                   },
    { PXS_SYS_OVERVIEW_SITE_NAME      , L"Site Name"                     },
    { PXS_SYS_OVERVIEW_ROLES          , L"Roles"                         },
    { PXS_SYS_OVERVIEW_DESCRIPTION    , L"Description"                   },
    { PXS_SYS_OVERVIEW_OPERATING_SYS  , L"Operating System"              },
    { PXS_SYS_OVERVIEW_MANUFACTURER   , L"Manufacturer"                  },
    { PXS_SYS_OVERVIEW_MODEL          , L"Model"                         },
    { PXS_SYS_OVERVIEW_SERIAL_NUMBER  , L"Serial Number"                 },
    { PXS_SYS_OVERVIEW_ASSET_TAG      , L"Asset Tag"                     },
    { PXS_SYS_OVERVIEW_NUM_OF_PROCS   , L"Number of Processors"          },
    { PXS_SYS_OVERVIEW_PROCESSOR_DESC , L"Processor Description"         },
    { PXS_SYS_OVERVIEW_TOTAL_MEMORY_MB, L"Total Memory"                  },
    { PXS_SYS_OVERVIEW_TOTAL_HDD      , L"Total Hard Drive"              },
    { PXS_SYS_OVERVIEW_DISPLAY        , L"Display"                       },
    { PXS_SYS_OVERVIEW_BIOS_VERSION   , L"BIOS Version"                  },
    { PXS_SYS_OVERVIEW_USER_ACCOUNT   , L"User Account"                  },
    { PXS_SYS_OVERVIEW_SYS_UPTIME_SEC , L"System Uptime"                 },
    { PXS_SYS_OVERVIEW_LOCAL_TIME     , L"Local Time"                    },
    { PXS_ACTIVE_SETUP_NAME           , L"Name"                          },
    { PXS_ACTIVE_SETUP_VERSION        , L"Version"                       },
    { PXS_ACTIVE_SETUP_INSTALLED      , L"Installed"                     },
    { PXS_INSTAL_PROGS_NAME           , L"Name"                          },
    { PXS_INSTAL_PROGS_VENDOR         , L"Vendor"                        },
    { PXS_INSTAL_PROGS_VERSION        , L"Version"                       },
    { PXS_INSTAL_PROGS_PRODUCT_LANG   , L"Product Language"              },
    { PXS_INSTAL_PROGS_INSTALL_DATE   , L"Install Date"                  },
    { PXS_INSTAL_PROGS_INSTALL_LOC    , L"Install Location"              },
    { PXS_INSTAL_PROGS_INSTALL_SRC    , L"Install Source"                },
    { PXS_INSTAL_PROGS_INSTALL_STATE  , L"Install State"                 },
    { PXS_INSTAL_PROGS_ASSIGN_TYPE    , L"Assignment Type"               },
    { PXS_INSTAL_PROGS_PACKAGE_CODE   , L"Package Code"                  },
    { PXS_INSTAL_PROGS_PACKAGE_NAME   , L"Package Name"                  },
    { PXS_INSTAL_PROGS_LOCAL_PACKAGE  , L"Local Package"                 },
    { PXS_INSTAL_PROGS_PRODUCT_ID     , L"Product ID"                    },
    { PXS_INSTAL_PROGS_REG_COMPANY    , L"Registered Company"            },
    { PXS_INSTAL_PROGS_REG_OWNER      , L"Registered Owner"              },
    { PXS_INSTAL_PROGS_TIMES_USED     , L"Times Used"                    },
    { PXS_INSTAL_PROGS_LAST_USED      , L"Last Used"                     },
    { PXS_INSTAL_PROGS_EXE_PATH       , L"Executable Path"               },
    { PXS_INSTAL_PROGS_EXE_VERSION    , L"Executable Version"            },
    { PXS_INSTAL_PROGS_EXE_DESCRIP    , L"Executable Description"        },
    { PXS_INSTAL_PROGS_SOFTWARE_ID    , L"Software ID"                   },
    { PXS_SOFTWARE_UPDATES_UPDATE_ID  , L"Update ID"                     },
    { PXS_SOFTWARE_UPDATES_INSTALL_ON , L"Installed On"                  },
    { PXS_SOFTWARE_UPDATES_DESCRIPTION, L"Description"                   },
    { PXS_OS_NAME                     , L"Name"                          },
    { PXS_OS_EDITION                  , L"Edition"                       },
    { PXS_OS_INSTALL_DATE             , L"Install Date"                  },
    { PXS_OS_REGISTERED_OWNER         , L"Registered Owner"              },
    { PXS_OS_REGISTERED_ORG           , L"Registered Organization"       },
    { PXS_OS_PRODUCT_ID               , L"Product ID"                    },
    { PXS_OS_MAJOR_VERSION_NUMBER     , L"Major Version Number"          },
    { PXS_OS_MINOR_VERSION_NUMBER     , L"Minor Version Number"          },
    { PXS_OS_BUILD_NUMBER             , L"Build Number"                  },
    { PXS_OS_CSD_VERSION              , L"Service Pack"                  },
    { PXS_OS_SERVICE_PACK_MAJOR_MINOR , L"Service Pack Version"          },
    { PXS_OS_PLUS_VERSION_NUMBER      , L"Plus! Version Number"          },
    { PXS_OS_DIRECTX_VERSION          , L"DirectX Version"               },
    { PXS_OS_WINDOWS_DIRECTORY        , L"Windows Directory"             },
    { PXS_OS_SYSTEM_DIRECTORY         , L"System Directory"              },
    { PXS_OS_TEMPORARY_DIRECTORY      , L"Temporary Directory"           },
    { PXS_OS_LANGUAGE                 , L"Operating System Language"     },
    { PXS_OS_NUMBER_OF_BITS           , L"Number of Bits"                },
    { PXS_OS_WIN10_VERSION            , L"Windows 10 Version"            },
    { PXS_OS_WIN10_OS_BUILD           , L"Windows 10 OS Build"           },
    { PXS_PERIPHERALS_NAME            , L"Name"                          },
    { PXS_PERIPHERALS_DESCRIPTION     , L"Description"                   },
    { PXS_INTERNET_SOFTWARE_TYPE      , L"Type"                          },
    { PXS_INTERNET_SOFTWARE_NAME      , L"Name"                          },
    { PXS_INTERNET_SOFTWARE_VERSION   , L"Version"                       },
    { PXS_INTERNET_SOFTWARE_DATA_UP   , L"Data Update"                   },
    { PXS_OPEN_PORTS_PROTOCOL         , L"Port Protocol"                 },
    { PXS_OPEN_PORTS_LOCAL_ADDRESS    , L"Local Address"                 },
    { PXS_OPEN_PORTS_LOCAL_PORT       , L"Local Port"                    },
    { PXS_OPEN_PORTS_CAPTION          , L"Caption"                       },
    { PXS_OPEN_PORTS_SERVICE_NAME     , L"Service Name"                  },
    { PXS_OPEN_PORTS_REMOTE_ADDRESS   , L"Remote Address"                },
    { PXS_OPEN_PORTS_REMOTE_PORT      , L"Remote Port"                   },
    { PXS_OPEN_PORTS_CONNECTION_STATE , L"Connection State"              },
    { PXS_OPEN_PORTS_PROCESS_NAME     , L"Process Name"                  },
    { PXS_OPEN_PORTS_PROCESS_ID       , L"Process ID"                    },
    { PXS_OPEN_PORTS_PROCESS_DESC     , L"Process Description"           },
    { PXS_OPEN_PORTS_PROCESS_MANUFAC  , L"Process Manufacturer"          },
    { PXS_SECURITY_LOG_TIME_GENERATED , L"Time Generated"                },
    { PXS_SECURITY_LOG_FILE           , L"Log File"                      },
    { PXS_SECURITY_LOG_SOURCE_NAME    , L"Source Name"                   },
    { PXS_SECURITY_LOG_DESCRIPTION    , L"Description"                   },
    { PXS_SECURITY_SETTINGS_ITEM      , L"Item"                          },
    { PXS_SECURITY_SETTINGS_NAME      , L"Name"                          },
    { PXS_SECURITY_SETTINGS_SETTING   , L"Setting"                       },
    { PXS_SYSRESTORE_SEQUENCE         , L"Sequence"                      },
    { PXS_SYSRESTORE_CREATE_TIME      , L"Creation Time"                 },
    { PXS_SYSRESTORE_DESCRIPTION      , L"Description"                   },
    { PXS_USERPRIVS_PRIVILEGE_NAME    , L"Privilege"                     },
    { PXS_WINDOWS_FIREWALL_NAME       , L"Name"                          },
    { PXS_WINDOWS_FIREWALL_SETTING    , L"Setting"                       },
    { PXS_GROUPS_TYPE                 , L"Group Type"                    },
    { PXS_GROUPS_NAME                 , L"Group Name"                    },
    { PXS_GROUPS_COMMENT              , L"Comment"                       },
    { PXS_GROUPMEMBERS_GROUP_NAME     , L"Group Name"                    },
    { PXS_GROUPMEMBERS_MEMBER_NAME    , L"Member Name"                   },
    { PXS_GROUPPOLICY_GROUP_NAME      , L"Group Name"                    },
    { PXS_GROUPPOLICY_PRIVILEGE_NAME  , L"Privilege Name"                },
    { PXS_USERS_USERNAME              , L"User Account"                  },
    { PXS_USERS_FULL_NAME             , L"Full Name"                     },
    { PXS_USERS_DESCRIPTION           , L"Description"                   },
    { PXS_USERS_ACCOUNT_STATUS        , L"Account Status"                },
    { PXS_USERS_LOCAL_GROUPS          , L"Local Groups"                  },
    { PXS_USERS_GLOBAL_GROUPS         , L"Global Groups"                 },
    { PXS_USERS_LAST_LOGON            , L"Last Logon"                    },
    { PXS_USERS_LAST_LOGOFF           , L"Last Logoff"                   },
    { PXS_USERS_NUMBER_OF_LOGONS      , L"Number of Logons"              },
    { PXS_USERS_BAD_PASSWORD_COUNT    , L"Bad Password Count"            },
    { PXS_USERS_PASSWORD_AGE_DAY      , L"Password Age"                  },
    { PXS_USERS_PASSWORD_EXPIRED      , L"Password Expired"              },
    { PXS_USERS_ACCOUNT_EXPIRES       , L"Account Expires"               },
    { PXS_SCHED_TASKS_TASK_NAME       , L"Task Name"                     },
    { PXS_SCHED_TASKS_STATUS          , L"Status"                        },
    { PXS_SCHED_TASKS_SCHEDULE        , L"Schedule"                      },
    { PXS_SCHED_TASKS_NEXT_RUN_TIME   , L"Next Run Time"                 },
    { PXS_SCHED_TASKS_RUN_COMMAND     , L"Run Command"                   },
    { PXS_SCHED_TASKS_MAX_RUN_TIME    , L"Maximum Run Time"              },
    { PXS_SCHED_TASKS_LAST_RUN_TIME   , L"Last Run Time"                 },
    { PXS_SCHED_TASKS_LAST_RESULT     , L"Last Result"                   },
    { PXS_UPTIME_DATA_START           , L"Data Start"                    },
    { PXS_UPTIME_SYSTEM_UPTIME        , L"System Uptime"                 },
    { PXS_UPTIME_SYSTEM_AVAIL_PERCENT , L"System Availability"           },
    { PXS_UPTIME_TOTAL_UPTIME         , L"Total Uptime"                  },
    { PXS_UPTIME_TOTAL_DOWNTIME       , L"Total Downtime"                },
    { PXS_UPTIME_TIMES_BOOTED         , L"Times Booted"                  },
    { PXS_UPTIME_CLEAN_SHUTDOWNS      , L"Clean Shutdowns"               },
    { PXS_UPTIME_UNEXPECTED_SHUTDOWNS , L"Unexpected Shutdowns"          },
    { PXS_ERROR_LOG_GENERATED         , L"Time Generated"                },
    { PXS_ERROR_LOG_FILE              , L"Log File"                      },
    { PXS_ERROR_LOG_SOURCE_NAME       , L"Source Name"                   },
    { PXS_ERROR_LOG_DESCRIPTION       , L"Description"                   },
    { PXS_WIN_NET_FILES_PATHNAME      , L"Path"                          },
    { PXS_WIN_NET_FILES_USERNAME      , L"User Account"                  },
    { PXS_WIN_NET_FILES_LOCKS         , L"Locks"                         },
    { PXS_WIN_NET_FILES_PERMISSIONS   , L"Permissions"                   },
    { PXS_WIN_NET_SESS_COMPUTERNAME   , L"Computer Name"                 },
    { PXS_WIN_NET_SESS_USERNAME       , L"User Account"                  },
    { PXS_WIN_NET_SESS_CONN_TIME_MIN  , L"Connected Time"                },
    { PXS_WIN_NET_SESS_IDLE_TIME_MIN  , L"Idle Time"                     },
    { PXS_WIN_NET_SHARES_NETNAME      , L"Share Name"                    },
    { PXS_WIN_NET_SHARES_TYPE         , L"Share Type"                    },
    { PXS_WIN_NET_SHARES_CONNECTIONS  , L"Connections"                   },
    { PXS_WIN_NET_SHARES_PATH         , L"Share Path"                    },
    { PXS_NET_ADAPT_ITEM_NUMBER       , L"Adapter Number"                },
    { PXS_NET_ADAPT_ADAPTER_NAME      , L"Adapter Name"                  },
    { PXS_NET_ADAPT_DNS_HOSTNAME      , L"DNS Host Name"                 },
    { PXS_NET_ADAPT_DNS_SERVERS       , L"DNS Servers"                   },
    { PXS_NET_ADAPT_IP_ADDRESS        , L"IP Address"                    },
    { PXS_NET_ADAPT_IP_SUBNET         , L"IP Subnet"                     },
    { PXS_NET_ADAPT_DEFAULT_GATEWAY   , L"Default IP Gateway"            },
    { PXS_NET_ADAPT_DHCP_ENABLED      , L"DHCP Enabled"                  },
    { PXS_NET_ADAPT_DHCP_SERVER       , L"DHCP Server"                   },
    { PXS_NET_ADAPT_DHCP_IPADDRESS    , L"DHCP IP Address"               },
    { PXS_NET_ADAPT_DHCP_LEASE_OBTAIN , L"DHCP Lease Obtained"           },
    { PXS_NET_ADAPT_DHCP_LEASE_EXPIRES, L"DHCP Lease Expires"            },
    { PXS_NET_ADAPT_STATUS_CODE       , L"Status Code"                   },
    { PXS_NET_ADAPT_ADAPTER_STATUS    , L"Adapter Status"                },
    { PXS_NET_ADAPT_ADAPTER_TYPE      , L"Adapter Type"                  },
    { PXS_NET_ADAPT_MAC_ADDRESS       , L"MAC Address"                   },
    { PXS_NET_ADAPT_CONN_STATUS       , L"Connection Status"             },
    { PXS_NET_ADAPT_CONNECT_MBPS      , L"Connection Speed"              },
    { PXS_NETBIOS_ITEM_NUMBER         , L"Adapter Number"                },
    { PXS_NETBIOS_SOFTWARE_RELEASE    , L"Software Release"              },
    { PXS_NETBIOS_MAC_ADDRESS         , L"MAC Address"                   },
    { PXS_NETBIOS_ADAPTER_TYPE        , L"Adapter Type"                  },
    { PXS_NETBIOS_MAX_SESSIONS        , L"Maximum Sessions"              },
    { PXS_NETBIOS_SESSIONS_PENDING    , L"Sessions Pending"              },
    { PXS_NETBIOS_MAX_DATA_BYTES      , L"Maximum Data"                  },
    { PXS_NETBIOS_MAX_DATAGRAM_BYTES  , L"Maximum Datagram"              },
    { PXS_HARDWARE_DEVS_DEVICE_TYPE   , L"Device Type"                   },
    { PXS_HARDWARE_DEVS_DEVICE_NAME   , L"Device Name"                   },
    { PXS_HARDWARE_DEVS_DESCRIPTION   , L"Description"                   },
    { PXS_HARDWARE_DEVS_MANUFACTURER  , L"Manufacturer"                  },
    { PXS_HARDWARE_DEVS_LOCATION      , L"Location"                      },
    { PXS_HARDWARE_DEVS_DRIVER_PROV   , L"Driver Provider"               },
    { PXS_HARDWARE_DEVS_DRIVER_VER    , L"Driver Version"                },
    { PXS_HARDWARE_DEVS_DRIVER_DATE   , L"Driver Date"                   },
    { PXS_HARDWARE_DEVS_STATUS_CODE   , L"Status Code"                   },
    { PXS_HARDWARE_DEVS_STATUS_MSG    , L"Status Message"                },
    { PXS_HARDWARE_DEVS_CLASS_GUID    , L"Class GUID"                    },
    { PXS_HARDWARE_DEVS_DEVICE_ID     , L"Device ID"                     },
    { PXS_DISPLAY_CAPS_DISPLAY_NUMBER , L"Display Number"                },
    { PXS_DISPLAY_CAPS_TYPE           , L"Type"                          },
    { PXS_DISPLAY_CAPS_DESCRIPTION    , L"Description"                   },
    { PXS_DISPLAY_CAPS_CAPABLE        , L"Capable"                       },
    { PXS_PRINTERS_ITEM_NUMBER        , L"Printer Number"                },
    { PXS_PRINTERS_NAME               , L"Printer Name"                  },
    { PXS_PRINTERS_SHARE_NAME         , L"Share Name"                    },
    { PXS_PRINTERS_PORT_NAME          , L"Port Name"                     },
    { PXS_PRINTERS_LOCATION           , L"Location"                      },
    { PXS_PRINTERS_AVERAGE_PPM        , L"Pages Per Minute"              },
    { PXS_PRINTERS_ATTRIBUTES         , L"Attributes"                    },
    { PXS_PRINTERS_PRINTER_STATUS     , L"Printer Status"                },
    { PXS_PRINTERS_PAPER_SIZE         , L"Paper Size"                    },
    { PXS_PRINTERS_ORIENTATION        , L"Orientation"                   },
    { PXS_PRINTERS_PRINT_QUALITY      , L"Print Quality"                 },
    { PXS_PRINTERS_DRIVER_NAME        , L"Driver Name"                   },
    { PXS_PRINTERS_DRIVER_FILE        , L"Driver File"                   },
    { PXS_PRINTERS_DRIVER_VERSION     , L"Driver Version"                },
    { PXS_PRINTERS_DRIVER_MANUFACTURER, L"Driver Manufacturer"           },
    { PXS_PRINTERS_DRIVER_DESCRIPTION , L"Driver Description"            },
    { PXS_PRINTERS_DRIVER_DATA_FILE   , L"Driver Data File"              },
    { PXS_PRINTERS_DRIVER_CONFIG_FILE , L"Driver Configuration File"     },
    { PXS_BIOS_VERSION_BIOS_VERSION   , L"BIOS Version"                  },
    { PXS_BIOS_VERSION_RELEASE_DATE   , L"Release Date"                  },
    { PXS_BIOS_VERSION_VIDEO_BIOS_VER , L"Video BIOS Version"            },
    { PXS_BIOS_VERSION_VIDEO_BIOS_DATE, L"Video BIOS Date"               },
    { PXS_CPU_BASIC_ITEM_NUMBER       , L"Processor Number"              },
    { PXS_CPU_BASIC_FULL_NAME         , L"Name"                          },
    { PXS_CPU_BASIC_SHORT_NAME        , L"Short Name"                    },
    { PXS_CPU_BASIC_SPEED_ESTIMATE_MHZ, L"Speed Estimated"               },
    { PXS_CPU_BASIC_SPEED_REGISTRY_MHZ, L"Speed Registry"                },
    { PXS_CPU_BASIC_TYPE              , L"Processor Type"                },
    { PXS_CPU_BASIC_MANUFACTURER      , L"Manufacturer"                  },
    { PXS_CPU_BASIC_SERIAL_NUMBER     , L"Serial Number"                 },
    { PXS_CPU_BASIC_APIC_PHYSICAL_ID  , L"APIC Physical ID"              },
    { PXS_CPU_BASIC_FEATURES          , L"Features"                      },
    { PXS_CPU_BASIC_CACHE             , L"Cache"                         },
    { PXS_CPU_BASIC_TLB               , L"TLB"                           },
    { PXS_CPU_BASIC_NUM_LOGICALS      , L"Logical Processors"            },
    { PXS_CPU_DETAILED_ITEM_NUMBER    , L"Processor Number"              },
    { PXS_CPU_DETAILED_NUM_LOGICAL    , L"Logical Processors"            },
    { PXS_CPU_DETAILED_NUM_CORES      , L"Number Cores"                  },
    { PXS_CPU_DETAILED_VENDOR_STRING  , L"Vendor String"                 },
    { PXS_CPU_DETAILED_FAMILY         , L"Family"                        },
    { PXS_CPU_DETAILED_MODEL          , L"Model"                         },
    { PXS_CPU_DETAILED_STEPPING       , L"Stepping"                      },
    { PXS_CPU_DETAILED_EXT_FAMILY     , L"Extended Family"               },
    { PXS_CPU_DETAILED_EXT_MODEL      , L"Extended Model"                },
    { PXS_CPU_DETAILED_BRAND_ID       , L"Brand ID"                      },
    { PXS_CPU_DETAILED_FPU            , L"FPU Floating Point Unit"       },
    { PXS_CPU_DETAILED_VME            , L"VME Virtual Mode Extensions"   },
    { PXS_CPU_DETAILED_DE             , L"DE Debugging Extensions"       },
    { PXS_CPU_DETAILED_PSE            , L"PSE Page Size Extensions"      },
    { PXS_CPU_DETAILED_TSC            , L"TSC Time Stamp Counter"        },
    { PXS_CPU_DETAILED_MSR            , L"MSR Model Specific Registers"  },
    { PXS_CPU_DETAILED_PAE            , L"PAE Physical Address Exts."    },
    { PXS_CPU_DETAILED_MCE            , L"MCE Machine Check Exceptions"  },
    { PXS_CPU_DETAILED_CX8            , L"CX8 Compare Exchange 8 Byte"   },
    { PXS_CPU_DETAILED_APIC           , L"APIC On Chip APIC present"     },
    { PXS_CPU_DETAILED_SEP            , L"SEP Fast System Calls"         },
    { PXS_CPU_DETAILED_MTRR           , L"MTRR Memory Type Range Regs."  },
    { PXS_CPU_DETAILED_PGE            , L"PGE Page Global Enable"        },
    { PXS_CPU_DETAILED_MCA            , L"MCA Machine Check Architecture"},
    { PXS_CPU_DETAILED_CMOV           , L"CMOV Conditional Move"         },
    { PXS_CPU_DETAILED_PAT            , L"PAT Page Attribute Table"      },
    { PXS_CPU_DETAILED_PSE36          , L"PSE36 36bit Page Size Exts"    },
    { PXS_CPU_DETAILED_PSN            , L"PSN Processor Serial Number"   },
    { PXS_CPU_DETAILED_CLFSH          , L"CLFSH CLFLUSH Instruction"     },
    { PXS_CPU_DETAILED_DS             , L"DS Debug Store"                },
    { PXS_CPU_DETAILED_ACPI           , L"ACPI ACPI Supported"           },
    { PXS_CPU_DETAILED_MMX            , L"MMX MMX Technology"            },
    { PXS_CPU_DETAILED_FXFR           , L"FXFR Fast Save Restore"        },
    { PXS_CPU_DETAILED_SSE            , L"SSE Streaming SIMD Exts"       },
    { PXS_CPU_DETAILED_SSE2           , L"SSE2 Streaming SIMD Exts 2"    },
    { PXS_CPU_DETAILED_SS             , L"SS Selfsnoop"                  },
    { PXS_CPU_DETAILED_HTT            , L"HTT Hyper Threading Technology"},
    { PXS_CPU_DETAILED_TM             , L"TM1 Thermal Interrupt Status"  },
    { PXS_CPU_DETAILED_IA64           , L"IA64 64 bit Intel"             },
    { PXS_CPU_DETAILED_PBE            , L"PBE Pending Break Event"       },
    { PXS_CPU_DETAILED_MMXEXT         , L"MMEXT MMX Technology"          },
    { PXS_CPU_DETAILED_3DNOW          , L"NOW3D Support 3DNow!"          },
    { PXS_CPU_DETAILED_3DNOWEXT       , L"NOW3DEXT Support 3DNowExt!+"   },
    { PXS_CPU_DETAILED_SSE3           , L"SSE3 Streaming SIMD Exts. 3"   },
    { PXS_CPU_DETAILED_EST            , L"EST Enhanced SpeedStep"        },
    { PXS_CPU_DETAILED_CX16           , L"CX16 Compare Exchange 16 Byte" },
    { PXS_CPU_CACHE_LEVEL             , L"Cache Level"                   },
    { PXS_CPU_CACHE_CPU_NUMBER        , L"Processor Number"              },
    { PXS_CPU_CACHE_IS_UNIFIED        , L"Unified Cache"                 },
    { PXS_CPU_CACHE_CODE_SIZE_KB      , L"Code Size"                     },
    { PXS_CPU_CACHE_CODE_WIDTH_BYTE   , L"Code Width"                    },
    { PXS_CPU_CACHE_CODE_ASSOC        , L"Code Associativity"            },
    { PXS_CPU_CACHE_DATA_SIZE_KB      , L"Data Size"                     },
    { PXS_CPU_CACHE_DATA_WIDTH_BYTE   , L"Data Width"                    },
    { PXS_CPU_CACHE_DATA_ASSOC        , L"Data Associativity"            },
    { PXS_CPU_CACHE_UNIFIED_SIZE_KB   , L"Unified Size"                  },
    { PXS_CPU_CACHE_UNIFIED_WIDTH_BYTE, L"Unified Width"                 },
    { PXS_CPU_CACHE_UNIFIED_ASSOC     , L"Unified Associativity"         },
    { PXS_CPU_TLB_CPU_NUMBER          , L"Processor Number"              },
    { PXS_CPU_TLB_IS_UNIFIED          , L"Unified Cache"                 },
    { PXS_CPU_TLB_4K_CODE_ENTRIES     , L"Code Entries 4K"               },
    { PXS_CPU_TLB_4K_CODE_ASSOC       , L"Code Associativity 4K"         },
    { PXS_CPU_TLB_4M_CODE_ENTRIES     , L"Code Entries 4M"               },
    { PXS_CPU_TLB_4M_CODE_ASSOC       , L"Code Associativity 4M"         },
    { PXS_CPU_TLB_4K_DATA_ENTRIES     , L"Data Entries 4K"               },
    { PXS_CPU_TLB_4K_DATA_ASSOC       , L"Data Associativity 4K"         },
    { PXS_CPU_TLB_4M_DATA_ENTRIES     , L"Data Entries 4M"               },
    { PXS_CPU_TLB_4M_DATA_ASSOC       , L"Data Associativity 4M"         },
    { PXS_CPU_TLB_4K_UNIFIED_ENTRIES  , L"Unified Entries 4K"            },
    { PXS_CPU_TLB_4K_UNIFIED_ASSOC    , L"Unified Associativity 4K"      },
    { PXS_CPU_TLB_4M_UNIFIED_ENTRIES  , L"Unified Entries 4M"            },
    { PXS_CPU_TLB_4M_UNIFIED_ASSOC    , L"Unified Associativity 4M"      },
    { PXS_MEMORY_TOTAL_RAM_MB         , L"Total Memory"                  },
    { PXS_MEMORY_FREE_RAM_MB          , L"Free Memory"                   },
    { PXS_MEMORY_MAX_SWAP_MB          , L"Maximum Swap File"             },
    { PXS_MEMORY_FREE_SWAP_MB         , L"Free Swap File"                },
    { PXS_PHYS_DISKS_ITEM_NUMBER      , L"Disk Number"                   },
    { PXS_PHYS_DISKS_CAPACITY_MB      , L"Capacity"                      },
    { PXS_PHYS_DISKS_TYPE             , L"Disk Type"                     },
    { PXS_PHYS_DISKS_MANUFACTURER     , L"Manufacturer"                  },
    { PXS_PHYS_DISKS_MODEL            , L"Model"                         },
    { PXS_PHYS_DISKS_SERIAL_NUMBER    , L"Serial Number"                 },
    { PXS_PHYS_DISKS_FIRMWARE_REV     , L"Firmware Revision"             },
    { PXS_PHYS_DISKS_CONTROL_RANK     , L"Controller Rank"               },
    { PXS_PHYS_DISKS_MASTER_SLAVE     , L"Master/Slave"                  },
    { PXS_PHYS_DISKS_TOTAL_CYLINDERS  , L"Total Cylinders"               },
    { PXS_PHYS_DISKS_TOTAL_HEADS      , L"Total Heads"                   },
    { PXS_PHYS_DISKS_SECTORS_PER_TRK  , L"Sectors Per Track"             },
    { PXS_PHYS_DISKS_BUFFER_KB        , L"Buffer Size"                   },
    { PXS_PHYS_DISKS_SMART_SUPPORTED  , L"SMART Supported"               },
    { PXS_PHYS_DISKS_SMART_ENABLED    , L"SMART Enabled"                 },
    { PXS_PHYS_DISKS_SMART_SELF_TEST  , L"SMART Self Test"               },
    { PXS_DRIVES_LETTER               , L"Letter"                        },
    { PXS_DRIVES_TYPE                 , L"Drive Type"                    },
    { PXS_DRIVES_USED_SPACE_PERCENT   , L"Percent Used"                  },
    { PXS_DRIVES_USED_SPACE           , L"Used Space"                    },
    { PXS_DRIVES_FREE_SPACE           , L"Free Space"                    },
    { PXS_DRIVES_TOTAL_SPACE          , L"Total Space"                   },
    { PXS_DRIVES_VOLUME_NAME          , L"Volume Name"                   },
    { PXS_DRIVES_FILE_SYSTEM          , L"File System"                   },
    { PXS_DRIVES_VOLUME_SERIAL_NUMBER , L"Volume Serial Number"          },
    { PXS_DRIVES_SECTORS_PER_CLUSTER  , L"Sectors Per Cluster"           },
    { PXS_DRIVES_BYTES_PER_SECTOR     , L"Bytes Per Sector"              },
    { PXS_DRIVES_FREE_CLUSTERS        , L"Free Clusters"                 },
    { PXS_DRIVES_TOTAL_CLUSTERS       , L"Total Clusters"                },
    { PXS_COMMPORTS_NUMBER            , L"Port Number"                   },
    { PXS_COMMPORTS_NAME              , L"Port Name"                     },
    { PXS_COMMPORTS_MONITOR_NAME      , L"Monitor Name"                  },
    { PXS_COMMPORTS_DESCRIPTION       , L"Description"                   },
    { PXS_COMMPORTS_TYPE              , L"Port Type"                     },
    { PXS_STARTUP_PROGRAM_NAME        , L"Program Name"                  },
    { PXS_STARTUP_SETTINGS_FOLDER     , L"Settings Folder"               },
    { PXS_STARTUP_COMMAND             , L"Command"                       },
    { PXS_NTSERVICES_SERVICE_NAME     , L"Name"                          },
    { PXS_NTSERVICES_TYPE             , L"Service Type"                  },
    { PXS_NTSERVICES_STATE            , L"State"                         },
    { PXS_NTSERVICES_START_MODE       , L"Start Mode"                    },
    { PXS_NTSERVICES_PATH_NAME        , L"Path Name"                     },
    { PXS_NTSERVICES_START_NAME       , L"Start Name"                    },
    { PXS_RUNNING_PROCS_NAME          , L"Name"                          },
    { PXS_RUNNING_PROCS_PROCESS_ID    , L"Process ID"                    },
    { PXS_RUNNING_PROCS_MEM_USAGE_KB  , L"Memory"                        },
    { PXS_RUNNING_PROCS_DESCRIPTION   , L"Description"                   },
    { PXS_LOADED_MODULES_NAME         , L"Name"                          },
    { PXS_LOADED_MODULES_VERSION      , L"Version"                       },
    { PXS_LOADED_MODULES_MODIFIED     , L"Modified"                      },
    { PXS_LOADED_MODULES_MANUFACTURER , L"Manufacturer"                  },
    { PXS_SYSTEM_FILES_NAME           , L"Name"                          },
    { PXS_SYSTEM_FILES_VERSION        , L"Version"                       },
    { PXS_SYSTEM_FILES_MODIFIED       , L"Modified"                      },
    { PXS_SYSTEM_FILES_MANUFACTURER   , L"Manufacturer"                  },
    { PXS_FIND_FILES_FILENAME         , L"Name"                          },
    { PXS_FIND_FILES_SIZE             , L"File Size"                     },
    { PXS_FIND_FILES_MOD_TIMESTAMP    , L"Modified"                      },
    { PXS_FIND_FILES_DIRECTORY        , L"Directory"                     },
    { PXS_DISPLAY_EDID_ITEM_NUMBER    , L"Display Number"                },
    { PXS_DISPLAY_EDID_DISPLAY_NAME   , L"Display Name"                  },
    { PXS_DISPLAY_EDID_MANUFACTURER   , L"Manufacturer"                  },
    { PXS_DISPLAY_EDID_MANUFAC_DATE   , L"Manufacture Date"              },
    { PXS_DISPLAY_EDID_SERIAL_NUMBER  , L"Serial Number"                 },
    { PXS_DISPLAY_EDID_PRODUCT_ID     , L"Product ID"                    },
    { PXS_DISPLAY_EDID_DISPLAY_SIZE   , L"Display Size"                  },
    { PXS_DISPLAY_EDID_DISPLAY_TYPE   , L"Display Type"                  },
    { PXS_DISPLAY_EDID_FEATURES       , L"Supported Features"            },
    { PXS_DISPLAY_RESOLUTION          , L"Display Resolution"            },
    { PXS_DISP_ADAPT_ITEM_NUMBER      , L"Adapter Number"                },
    { PXS_DISP_ADAPT_NAME             , L"Name"                          },
    { PXS_DISP_ADAPT_ADAPTER_RAM_MB   , L"Adapter RAM"                   },
    { PXS_DISP_ADAPT_COLOR_DEPTH_BIT  , L"Colour Depth"                  },
    { PXS_DISP_ADAPT_LOGPIXELSY_DPI   , L"Vertical Resolution"           },
    { PXS_DISP_ADAPT_CURR_REFRESH_HZ  , L"Current Refresh Rate"          },
    { PXS_DISP_ADAPT_VIDEO_PROCESSOR  , L"Video Processor"               },
    { PXS_DISP_ADAPT_ADAPTER_DAC_TYPE , L"Adapter DAC Type"              },
    { PXS_DISP_ADAPT_ADAPTER_ID       , L"Adapter ID"                    },
    { PXS_DISP_ADAPT_BIOS             , L"BIOS"                          },
    { PXS_ENVIRON_VARS_NAME           , L"Name"                          },
    { PXS_ENVIRON_VARS_VALUE          , L"Variable Value"                },
    { PXS_REGION_SETTINGS_ITEM        , L"Item"                          },
    { PXS_REGION_SETTINGS_NAME        , L"Name"                          },
    { PXS_REGION_SETTINGS_SETTING     , L"Setting"                       },
    { PXS_PERMISSIONS_OBJECT_NAME     , L"Name"                          },
    { PXS_PERMISSIONS_OBJECT_TYPE     , L"Object Type"                   },
    { PXS_PERMISSIONS_TRUSTEE         , L"Trustee"                       },
    { PXS_PERMISSIONS_ACE_TYPE        , L"ACE Type"                      },
    { PXS_PERMISSIONS_PERMISSIONS     , L"Permissions"                   },
    { PXS_PERMISSIONS_ACE_FLAGS       , L"ACE Flags"                     },
    { PXS_PERMISSIONS_ACCESS_MASK     , L"Access Mask"                   },
    { PXS_PERMISSIONS_OBJECT_OWNER    , L"Owner"                         },
    { PXS_USER_LOGONS_USER_NAME       , L"User Account"                  },
    { PXS_USER_LOGONS_FIRST_LOGON_TS  , L"First Logon Timestamp"         },
    { PXS_USER_LOGONS_LAST_LOGON_TS   , L"Last Logon Timestamp"          },
    { PXS_USER_LOGONS_CONSOLE_LOGONS  , L"Console Logons"                },
    { PXS_USER_LOGONS_REMOTE_LOGONS   , L"Remote Logons"                 },
    { PXS_USER_LOGONS_OTHER_LOGONS    , L"Other Logons"                  },
    { PXS_SOFT_METER_FILE_NAME        , L"File Name"                     },
    { PXS_SOFT_METER_FILE_PATH        , L"File Path"                     },
    { PXS_SOFT_METER_FILE_VERSION     , L"File Version"                  },
    { PXS_SOFT_METER_PUBLISHER        , L"Publisher"                     },
    { PXS_SOFT_METER_FIRST_START_TS   , L"First Start Timestamp"         },
    { PXS_SOFT_METER_LAST_START_TS    , L"Last Start Timestamp"          },
    { PXS_SOFT_METER_NUMBER_STARTS    , L"Number of Starts"              },
    { PXS_SOFT_METER_REMOTE_STARTS    , L"Remote Starts"                 },
    { PXS_SOFT_METER_OTHER_STARTS     , L"Other Starts"                  },
    { PXS_ODBC_DRIVERS_NAME           , L"Name"                          },
    { PXS_ODBC_DRIVERS_COMPANY        , L"Company"                       },
    { PXS_ODBC_DRIVERS_FILE_NAME      , L"File Name"                     },
    { PXS_ODBC_DRIVERS_FILE_VERSION   , L"File Version"                  },
    { PXS_ODBC_DRIVERS_DRIVER_ODBC_VER, L"Driver ODBC Version"           },
    { PXS_ODBC_DATA_SOURCES_DSN_TYPE  , L"DSN Type"                      },
    { PXS_ODBC_DATA_SOURCES_DSN       , L"DSN"                           },
    { PXS_ODBC_DATA_SOURCES_DESC      , L"Description"                   },
    { PXS_OTHER_AUDIT_DATA_ITEM_CLASS , L"Item Class"                    },
    { PXS_OTHER_AUDIT_DATA_ITEM_NAME  , L"Item Name"                     },
    { PXS_OTHER_AUDIT_DATA_ITEM_VALUE , L"Item Value"                    },
    { PXS_NON_WINDOWS_EXE_FILENAME    , L"Name"                          },
    { PXS_NON_WINDOWS_EXE_VERSION     , L"Version"                       },
    { PXS_NON_WINDOWS_EXE_MODIFIED    , L"Modified"                      },
    { PXS_NON_WINDOWS_EXE_MANUFAC     , L"Manufacturer"                  },
    { PXS_NON_WINDOWS_EXE_DIRECTORY   , L"Directory"                     },
    { PXS_SMBIOS_BIOS_SMBIOS_VERSION  , L"SMBIOS Version"                },
    { PXS_SMBIOS_BOARD_ITEM_NUMBER    , L"Board Number"                  },
    { PXS_SMBIOS_PROCESSOR_ITEM_NUMBER, L"Processor Number"              },
    { PXS_SMBIOS_MEMCTRL_ITEM_NUMBER  , L"Controller Number"             },
    { PXS_SMBIOS_MEMMODULE_ITEM_NUMBER, L"Module Number"                 },
    { PXS_SMBIOS_CPUCACHE_ITEM_NUMBER , L"Cache Number"                  },
    { PXS_SMBIOS_PORTCONN_ITEM_NUMBER , L"Port Number"                   },
    { PXS_SMBIOS_SYSSLOT_ITEM_NUMBER  , L"Slot Number"                   },
    { PXS_SMBIOS_MEMARRAY_ITEM_NUMBER , L"Array Number"                  },
    { PXS_SMBIOS_MEM_DEV_ITEM_NUMBER  , L"Device Number"                 },
    { PXS_NET_TIME_PROTOCOL_NAME      , L"Name"                          },
    { PXS_NET_TIME_PROTOCOL_SETTING   , L"Setting"                       },
    { PXS_USER_RIGHTS_POLICY          , L"Policy"                        },
    { PXS_USER_RIGHTS_SETTING         , L"Security Setting"              },
    { PXS_KERBEROS_POLICY_POLICY      , L"Policy"                        },
    { PXS_KERBEROS_POLICY_SETTING     , L"Security Setting"              },
    { PXS_KERBEROS_TICKS_SERVER_NAME  , L"Server Name"                   },
    { PXS_KERBEROS_TICKS_REALM_NAME   , L"Realm Name"                    },
    { PXS_KERBEROS_TICKS_START_TIME   , L"Start Time"                    },
    { PXS_KERBEROS_TICKS_END_TIME     , L"End Time"                      },
    { PXS_KERBEROS_TICKS_RENEW_TIME   , L"Renew Time"                    },
    { PXS_KERBEROS_TICKS_ENCRYPT_TYPE , L"Encryption Type"               },
    { PXS_KERBEROS_TICKS_TICKET_FLAGS , L"Ticket Flags"                  },
    { PXS_REG_SEC_VALS_SUBKEY_SUBKEY  , L"Subkey"                        },
    { PXS_REG_SEC_VALS_SUBKEY_SETTING , L"Setting"                       },
    { PXS_OLE_DB_PROVIDERS_NAME       , L"Name"                          },
    { PXS_OLE_DB_PROVIDERS_DESCRIPTION, L"Description"                   },
    { PXS_ROUTING_TABLE_DESTINATION   , L"Destination"                   },
    { PXS_ROUTING_TABLE_NETMASK       , L"Netmask"                       },
    { PXS_ROUTING_TABLE_NEXT_HOP      , L"Next Hop"                      },
    { PXS_ROUTING_TABLE_INTERFACE     , L"Interface"                     },
    { PXS_ROUTING_TABLE_ROUTE_TYPE    , L"Route Type"                    },
    { PXS_ROUTING_TABLE_PROTOCOL      , L"Protocol"                      },
    { PXS_ROUTING_TABLE_AGE           , L"Age"                           },
    { PXS_ROUTING_TABLE_METIC         , L"Metric"                        }
};

///////////////////////////////////////////////////////////////////////////////////////////////////
// Global POD Variables
///////////////////////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////////////////////////
// Global Functions
///////////////////////////////////////////////////////////////////////////////////////////////////

// Searching/Sorting
void PXSGetAuditItemDisplayName( DWORD itemID, String* pDisplayName );
int  PXSQSortNumberStringAscending( const void* pArg1, const void* pArg2 );
int  PXSBSearchCompareAuditItems( const void* pItem1, const void* pItem2 );
int  PXSBSearchComparePortServices( const void* pPort1, const void* pPort2);
void PXSSortAuditRecords( TArray< AuditRecord >* pRecords, DWORD itemID );

// Audit
void PXSAuditRecordsToContent( const TArray< AuditRecord >& AuditRecords,
                               TArray< TreeViewItem >* pCategoryItems,
                               DWORD* pTableCounter, String* pRichText );
void PXSAuditRecordsToCsv( const TArray< AuditRecord >&  AuditRecords,
                           bool wantHeaderRow, String* pCsvText );
void PXSAuditRecordsToCsv2( const TArray< AuditRecord >&  AuditRecords,
                            bool wantHeaderRow, String* pCsvText );
void PXSAuditRecordsToHtml( const TArray< AuditRecord >& AuditRecords, String* pHtmlText );
void PXSGetCommandLineSwitchValues( const StringArray& Switches,
                                    String* pReportSwitchValue,
                                    String* pFileSwitchValue,
                                    String* pLogSwitchValue,
                                    String* pTimestampSwitchValue, String* pLanguageSwitchValue );
void PXSGetDataCategoryProperties( DWORD categoryID,
                                   String* pCategoryName,
                                   DWORD*  pCaptionID,
                                   bool*   pIsColumnar, bool* pIsNode, BYTE* pDepth );
void PXSGetWinAuditGuidFilePath( String* pFilePath );
void PXSMakeCommandLineOutputPath( const String& FileSwitchValue,
                                   const String& TimestampSwitchValue,
                                   const String& LocalTimeIso, String* pOutputPath );
void PXSReadWinAuditGuidFile( String* pWinAuditGuid );
void PXSSaveAuditCommandLine( const String& OutputPath, const TArray<AuditRecord>& AuditRecords );
void PXSWriteWinAuditGuidFile();

#endif  // WINAUDIT_WINAUDIT_H_
